import pandas as pd
from collections import Counter, defaultdict
from os import path, mkdir
import altair as alt
import networkx as nx
from itertools import tee
import re
import numpy as np
from datetime import datetime

from mag_annotator.utils import get_database_locs, get_ids_from_annotation

# TODO: add RBH information to output
# TODO: add flag to output table and not xlsx
# TODO: add flag to output heatmap table

FRAME_COLUMNS = ['gene_id', 'gene_description', 'module', 'sheet', 'header', 'subheader']
RRNA_TYPES = ['5S rRNA', '16S rRNA', '23S rRNA']
HEATMAP_MODULES = ['M00001', 'M00004', 'M00008', 'M00009', 'M00012', 'M00165', 'M00173', 'M00374', 'M00375',
                   'M00376', 'M00377', 'M00422', 'M00567']
HEATMAP_CELL_HEIGHT = 10
HEATMAP_CELL_WIDTH = 10
KO_REGEX = r'^K\d\d\d\d\d$'
ETC_COVERAGE_COLUMNS = ['module_id', 'module_name', 'complex', 'genome', 'path_length', 'path_length_coverage',
                        'percent_coverage', 'genes', 'missing_genes', 'complex_module_name']
GENOMES_PER_LIQUOR = 1000
TAXONOMY_LEVELS = ['d', 'p', 'c', 'o', 'f', 'g', 's']


def get_ordered_uniques(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def fill_genome_summary_frame(annotations, genome_summary_frame, groupby_column):
    for genome, frame in annotations.groupby(groupby_column, sort=False):
        genome_summary_id_sets = [set([k.strip() for k in j.split(',')]) for j in genome_summary_frame['gene_id']]
        id_dict = get_ids_from_annotation(frame)
        counts = list()
        for i in genome_summary_id_sets:
            identifier_count = 0
            for j in i:
                if j in id_dict:
                    identifier_count += id_dict[j]
            counts.append(identifier_count)
        genome_summary_frame[genome] = counts
    return genome_summary_frame


def summarize_rrnas(rrnas_df, groupby_column='fasta'):
    genome_rrna_dict = dict()
    for genome, frame in rrnas_df.groupby(groupby_column):
        genome_rrna_dict[genome] = Counter(frame['type'])
    row_list = list()
    for rna_type in RRNA_TYPES:
        row = [rna_type, '%s ribosomal RNA gene' % rna_type.split()[0], 'rRNA', 'rRNA', '', '']
        for genome, rrna_dict in genome_rrna_dict.items():
            row.append(genome_rrna_dict[genome].get(rna_type, 0))
        row_list.append(row)
    rrna_frame = pd.DataFrame(row_list, columns=FRAME_COLUMNS + list(genome_rrna_dict.keys()))
    return rrna_frame


def summarize_trnas(trnas_df, groupby_column='fasta'):
    # first build the frame
    combos = {(line.Type, line.Codon, line.Note) for _, line in trnas_df.iterrows()}
    frame_rows = list()
    for combo in combos:
        if combo[2] == 'pseudo':
            gene_id = '%s, pseudo (%s)'
            gene_description = '%s pseudo tRNA with %s Codon'
        else:
            gene_id = '%s (%s)'
            gene_description = '%s tRNA with %s Codon'
        gene_id = gene_id % (combo[0], combo[1])
        gene_description = gene_description % (combo[0], combo[1])
        module_description = '%s tRNA' % combo[0]
        frame_rows.append([gene_id, gene_description, module_description, 'tRNA', 'tRNA', ''])
    trna_frame = pd.DataFrame(frame_rows, columns=FRAME_COLUMNS)
    trna_frame = trna_frame.sort_values('gene_id')
    # then fill it in
    trna_frame = trna_frame.set_index('gene_id')
    for group, frame in trnas_df.groupby(groupby_column):
        gene_ids = list()
        for index, line in frame.iterrows():
            if line.Note == 'pseudo':
                gene_id = '%s, pseudo (%s)'
            else:
                gene_id = '%s (%s)'
            gene_ids.append(gene_id % (line.Type, line.Codon))
        trna_frame[group] = pd.Series(Counter(gene_ids))
    trna_frame = trna_frame.reset_index()
    trna_frame = trna_frame.fillna(0)
    return trna_frame


def make_genome_summary(annotations, genome_summary_frame, trna_frame=None, rrna_frame=None,
                        groupby_column='fasta'):
    summary_frames = list()
    # get ko summaries
    summary_frames.append(fill_genome_summary_frame(annotations, genome_summary_frame.copy(), groupby_column))

    # add rRNAs
    if rrna_frame is not None:
        summary_frames.append(summarize_rrnas(rrna_frame, groupby_column))

    # add tRNAs
    if trna_frame is not None:
        summary_frames.append(summarize_trnas(trna_frame, groupby_column))

    # merge summary frames
    summarized_genomes = pd.concat(summary_frames, sort=False)
    return summarized_genomes


def write_summarized_genomes_to_xlsx(summarized_genomes, output_file):
    # turn all this into an xlsx
    with pd.ExcelWriter(output_file) as writer:
        for sheet, frame in summarized_genomes.groupby('sheet', sort=False):
            frame = frame.sort_values(['header', 'subheader', 'module', 'gene_id'])
            frame = frame.drop(['sheet'], axis=1)
            frame.to_excel(writer, sheet_name=sheet, index=False)


# TODO: add assembly stats like N50, longest contig, total assembled length etc
def make_genome_stats(annotations, rrna_frame=None, trna_frame=None, groupby_column='fasta'):
    rows = list()
    columns = ['genome']
    if 'scaffold' in annotations.columns:
        columns.append('number of scaffolds')
    if 'bin_taxonomy' in annotations.columns:
        columns.append('taxonomy')
    if 'bin_completeness' in annotations.columns:
        columns.append('completeness score')
    if 'bin_contamination' in annotations.columns:
        columns.append('contamination score')
    if rrna_frame is not None:
        columns += RRNA_TYPES
    if trna_frame is not None:
        columns.append('tRNA count')
    if 'bin_completeness' in annotations.columns and 'bin_contamination' in annotations.columns \
       and rrna_frame is not None and trna_frame is not None:
        columns.append('assembly quality')
    for genome, frame in annotations.groupby(groupby_column, sort=False):
        row = [genome]
        if 'scaffold' in frame.columns:
            row.append(len(set(frame['scaffold'])))
        if 'bin_taxonomy' in frame.columns:
            row.append(frame['bin_taxonomy'][0])
        if 'bin_completeness' in frame.columns:
            row.append(frame['bin_completeness'][0])
        if 'bin_contamination' in frame.columns:
            row.append(frame['bin_contamination'][0])
        has_rrna = list()
        if rrna_frame is not None:
            genome_rrnas = rrna_frame.loc[rrna_frame.fasta == genome]
            for rrna in RRNA_TYPES:
                sixteens = genome_rrnas.loc[genome_rrnas.type == rrna]
                if sixteens.shape[0] == 0:
                    row.append('')
                    has_rrna.append(False)
                elif sixteens.shape[0] == 1:
                    row.append('%s, (%s, %s)' % (sixteens[groupby_column].iloc[0], sixteens.begin.iloc[0],
                                                 sixteens.end.iloc[0]))
                    has_rrna.append(True)
                else:
                    row.append('%s present' % sixteens.shape[0])
                    has_rrna.append(False)
        if trna_frame is not None:
            row.append(trna_frame.loc[trna_frame[groupby_column] == genome].shape[0])  # TODO: remove psuedo from count?
        if 'assembly quality' in columns:
            if frame['bin_completeness'][0] > 90 and frame['bin_contamination'][0] < 5 and np.all(has_rrna) and \
               len(set(trna_frame.loc[trna_frame[groupby_column] == genome].Type)) >= 18:
                row.append('high')
            elif frame['bin_completeness'][0] >= 50 and frame['bin_contamination'][0] < 10:
                row.append('med')
            else:
                row.append('low')
        rows.append(row)
    genome_stats = pd.DataFrame(rows, columns=columns)
    return genome_stats


def build_module_net(module_df):
    """Starts with a data from including a single module"""
    # build net from a set of module paths
    num_steps = max([int(i.split(',')[0]) for i in set(module_df['path'])])
    module_net = nx.DiGraph(num_steps=num_steps, module_id=list(module_df['module'])[0],
                            module_name=list(module_df['module_name'])[0])
    # go through all path/step combinations
    for module_path, frame in module_df.groupby('path'):
        split_path = [int(i) for i in module_path.split(',')]
        step = split_path[0]
        module_net.add_node(module_path, kos=set(frame['ko']))
        # add incoming edge
        if step != 0:
            module_net.add_edge('end_step_%s' % (step - 1), module_path)
        # add outgoing edge
        module_net.add_edge(module_path, 'end_step_%s' % step)
    return module_net


def get_module_step_coverage(kos, module_net):
    # prune network based on what kos were observed
    pruned_module_net = module_net.copy()
    module_kos_present = set()
    for node, data in module_net.nodes.items():
        if 'kos' in data:
            ko_overlap = data['kos'] & kos
            if len(ko_overlap) == 0:
                pruned_module_net.remove_node(node)
            else:
                module_kos_present = module_kos_present | ko_overlap
    # count number of missing steps
    missing_steps = 0
    for node, data in pruned_module_net.nodes.items():
        if ('end_step' in node) and (pruned_module_net.in_degree(node) == 0):
            missing_steps += 1
    # get statistics
    num_steps = pruned_module_net.graph['num_steps'] + 1
    num_steps_present = num_steps - missing_steps
    coverage = num_steps_present / num_steps
    return num_steps, num_steps_present, coverage, sorted(module_kos_present)


def make_module_coverage_df(annotation_df, module_nets):
    kos_to_genes = defaultdict(list)
    for gene_id, ko_list in annotation_df['kegg_id'].iteritems():
        if type(ko_list) is str:
            for ko in ko_list.split(','):
                kos_to_genes[ko].append(gene_id)
    coverage_dict = {}
    for i, (module, net) in enumerate(module_nets.items()):
        module_steps, module_steps_present, module_coverage, module_kos = \
            get_module_step_coverage(set(kos_to_genes.keys()), net)
        module_genes = sorted([gene for ko in module_kos for gene in kos_to_genes[ko]])
        coverage_dict[module] = [net.graph['module_name'], module_steps, module_steps_present, module_coverage,
                                 len(module_kos), ','.join(module_kos), ','.join(module_genes)]
    coverage_df = pd.DataFrame.from_dict(coverage_dict, orient='index',
                                         columns=['module_name', 'steps', 'steps_present', 'step_coverage', 'ko_count',
                                                  'kos_present', 'genes_present'])
    return coverage_df


def make_module_coverage_frame(annotations, module_nets, groupby_column='fasta'):
    # go through each scaffold to check for modules
    module_coverage_dict = dict()
    for group, frame in annotations.groupby(groupby_column, sort=False):
        module_coverage_dict[group] = make_module_coverage_df(frame, module_nets)
    module_coverage = pd.concat(module_coverage_dict)
    module_coverage.index = module_coverage.index.set_names(['genome', 'module'])
    return module_coverage.reset_index()


def make_module_coverage_heatmap(module_coverage, mag_order=None):
    num_mags_in_frame = len(set(module_coverage['genome']))
    c = alt.Chart(module_coverage, title='Module').encode(
        x=alt.X('module_name', title=None, sort=None, axis=alt.Axis(labelLimit=0, labelAngle=90)),
        y=alt.Y('genome', title=None, sort=mag_order, axis=alt.Axis(labelLimit=0)),
        tooltip=[alt.Tooltip('genome', title='Genome'),
                 alt.Tooltip('module_name', title='Module Name'),
                 alt.Tooltip('steps', title='Module steps'),
                 alt.Tooltip('steps_present', title='Steps present')
                 ]
    ).mark_rect().encode(color=alt.Color('step_coverage', legend=alt.Legend(title='% Complete'),
                                         scale=alt.Scale(domain=(0, 1)))).properties(
        width=HEATMAP_CELL_WIDTH * len(HEATMAP_MODULES),
        height=HEATMAP_CELL_HEIGHT * num_mags_in_frame)
    return c


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def first_open_paren_is_all(str_):
    """Go through string and return true"""
    curr_level = 1
    for char in str_[1:-1]:
        if char == ')':
            curr_level -= 1
        elif char == '(':
            curr_level += 1
        if curr_level == 0:
            return False
    return True


def split_into_steps(definition, split_char=' '):
    """Very fancy split on string of chars"""
    curr_level = 0
    step_starts = [-1]
    for i, char in enumerate(definition):
        if char is '(':
            curr_level += 1
        if char is ')':
            curr_level -= 1
        if (curr_level == 0) and (char in split_char):
            step_starts.append(i)
    step_starts.append(len(definition))
    steps = list()
    for a, b in pairwise(step_starts):
        step = definition[a+1:b]
        if step.startswith('(') and step.endswith(')'):
            if first_open_paren_is_all(step):
                step = step[1:-1]
        steps.append(step)
    return steps


def is_ko(ko):
    return re.match(KO_REGEX, ko) is not None


def make_module_network(definition, network: nx.DiGraph = None, parent_nodes=('start',)):
    # TODO: Figure out how to add 'end' step to last step at end
    if network is None:
        network = nx.DiGraph()
    last_steps = []
    for step in split_into_steps(definition, ','):
        prev_steps = parent_nodes
        for substep in split_into_steps(step, '+'):
            if is_ko(substep):
                for prev_step in prev_steps:
                    network.add_edge(prev_step, substep)
                prev_steps = [substep]
            else:
                network, prev_steps = make_module_network(substep, network, prev_steps)
        last_steps += prev_steps
    return network, last_steps


def get_module_coverage(module_net: nx.DiGraph, genes_present: set):
    max_coverage = -1
    max_coverage_genes = list()
    max_coverage_missing_genes = list()
    max_path_len = 0
    for net_path in nx.all_simple_paths(module_net, source='start', target='end'):
        net_path = set(net_path[1:-1])
        overlap = net_path & genes_present
        coverage = len(overlap) / len(net_path)
        if coverage > max_coverage:
            max_coverage = coverage
            max_coverage_genes = overlap
            max_coverage_missing_genes = net_path - genes_present
            max_path_len = len(net_path)
    return max_path_len, len(max_coverage_genes), max_coverage, max_coverage_genes, max_coverage_missing_genes


def make_etc_coverage_df(etc_module_df, annotations, groupby_column='fasta'):
    etc_coverage_df_rows = list()
    for _, module_row in etc_module_df.iterrows():
        definition = module_row['definition']
        # remove optional subunits
        definition = re.sub(r'-K\d\d\d\d\d', '', definition)
        module_net, _ = make_module_network(definition)
        # add end node
        no_out = [node for node in module_net.nodes() if module_net.out_degree(node) == 0]
        for node in no_out:
            module_net.add_edge(node, 'end')
        # go through each genome and check pathway coverage
        for group, frame in annotations.groupby(groupby_column):
            # get annotation genes
            grouped_ids = set(get_ids_from_annotation(frame).keys())
            path_len, path_coverage_count, path_coverage_percent, genes, missing_genes = \
                get_module_coverage(module_net, grouped_ids)
            complex_module_name = 'Complex %s: %s' % (module_row['complex'].replace('Complex ', ''),
                                                      module_row['module_name'])
            etc_coverage_df_rows.append([module_row['module_id'], module_row['module_name'],
                                         module_row['complex'].replace('Complex ', ''), group, path_len,
                                         path_coverage_count, path_coverage_percent, ','.join(sorted(genes)),
                                         ','.join(sorted(missing_genes)), complex_module_name])
    return pd.DataFrame(etc_coverage_df_rows, columns=ETC_COVERAGE_COLUMNS)


def make_etc_coverage_heatmap(etc_coverage, mag_order=None, module_order=None):
    num_mags_in_frame = len(set(etc_coverage['genome']))
    charts = list()
    for i, (etc_complex, frame) in enumerate(etc_coverage.groupby('complex')):
        # if this is the first chart then make y-ticks otherwise none
        c = alt.Chart(frame, title=etc_complex).encode(
            x=alt.X('module_name', title=None, axis=alt.Axis(labelLimit=0, labelAngle=90),
                    sort=module_order),
            y=alt.Y('genome', axis=alt.Axis(title=None, labels=False, ticks=False), sort=mag_order),
            tooltip=[alt.Tooltip('genome', title='Genome'),
                     alt.Tooltip('module_name', title='Module Name'),
                     alt.Tooltip('path_length', title='Module Subunits'),
                     alt.Tooltip('path_length_coverage', title='Subunits present'),
                     alt.Tooltip('genes', title='Genes present'),
                     alt.Tooltip('missing_genes', title='Genes missing')
                     ]
        ).mark_rect().encode(color=alt.Color('percent_coverage', legend=alt.Legend(title='% Complete'),
                                             scale=alt.Scale(domain=(0, 1)))).properties(
            width=HEATMAP_CELL_WIDTH * len(set(frame['module_name'])),
            height=HEATMAP_CELL_HEIGHT * num_mags_in_frame)
        charts.append(c)
    concat_title = alt.TitleParams('ETC Complexes', anchor='middle')
    return alt.hconcat(*charts, spacing=5, title=concat_title)


def make_functional_df(annotations, function_heatmap_form, groupby_column='fasta'):
    # clean up function heatmap form
    function_heatmap_form = function_heatmap_form.apply(lambda x: x.str.strip() if x.dtype == "object" else x)
    function_heatmap_form = function_heatmap_form.fillna('')
    # build dict of ids per genome
    genome_to_id_dict = dict()
    for genome, frame in annotations.groupby(groupby_column, sort=False):
        id_list = get_ids_from_annotation(frame).keys()
        genome_to_id_dict[genome] = set(id_list)
    # build long from data frame
    rows = list()
    for function, frame in function_heatmap_form.groupby('function_name', sort=False):
        for bin_name, id_set in genome_to_id_dict.items():
            presents_in_bin = list()
            functions_present = set()
            for _, row in frame.iterrows():
                function_id_set = set([i.strip() for i in row.function_ids.strip().split(',')])
                present_in_bin = id_set & function_id_set
                functions_present = functions_present | present_in_bin
                presents_in_bin.append(len(present_in_bin) > 0)
            function_in_bin = np.all(presents_in_bin)
            row = frame.iloc[0]
            rows.append([row.category, row.subcategory, row.function_name, ', '.join(functions_present),
                         '; '.join(get_ordered_uniques(frame.long_function_name)),
                         '; '.join(get_ordered_uniques(frame.gene_symbol)), bin_name, function_in_bin,
                         '%s: %s' % (row.category, row.function_name)])
    return pd.DataFrame(rows, columns=list(function_heatmap_form.columns) + ['genome', 'present',
                                                                             'category_function_name'])


def make_functional_heatmap(functional_df, mag_order=None):
    # build heatmaps
    charts = list()
    for i, (group, frame) in enumerate(functional_df.groupby('category', sort=False)):
        # set variables for chart
        function_order = get_ordered_uniques(list(frame.function_name))
        num_mags_in_frame = len(set(frame['genome']))
        chart_width = HEATMAP_CELL_WIDTH * len(function_order)
        chart_height = HEATMAP_CELL_HEIGHT * num_mags_in_frame
        # if this is the first chart then make y-ticks otherwise none
        if i == 0:
            y = alt.Y('genome', title=None, sort=mag_order,
                      axis=alt.Axis(labelLimit=0, labelExpr="replace(datum.label, /_\d*$/gi, '')"))
        else:
            y = alt.Y('genome', axis=alt.Axis(title=None, labels=False, ticks=False), sort=mag_order)
        # set up colors for chart
        rect_colors = alt.Color('present',
                                legend=alt.Legend(title="Function is Present", symbolType='square',
                                                  values=[True, False]),
                                sort=[True, False],
                                scale=alt.Scale(range=['#2ca25f', '#e5f5f9']))
        # define chart
        # TODO: Figure out how to angle title to take up less space
        c = alt.Chart(frame, title=alt.TitleParams(group)).encode(
            x=alt.X('function_name', title=None, axis=alt.Axis(labelLimit=0, labelAngle=90), sort=function_order),
            tooltip=[alt.Tooltip('genome', title='Genome'),
                     alt.Tooltip('category', title='Category'),
                     alt.Tooltip('subcategory', title='Subcategory'),
                     alt.Tooltip('function_ids', title='Function IDs'),
                     alt.Tooltip('function_name', title='Function'),
                     alt.Tooltip('long_function_name', title='Description'),
                     alt.Tooltip('gene_symbol', title='Gene Symbol')]
        ).mark_rect().encode(y=y, color=rect_colors).properties(
            width=chart_width,
            height=chart_height)
        charts.append(c)
    # merge and return
    function_heatmap = alt.hconcat(*charts, spacing=5)
    return function_heatmap


# TODO: refactor this to handle splitting large numbers of genomes into multiple heatmaps here
def fill_liquor_dfs(annotations, module_nets, etc_module_df, function_heatmap_form, groupby_column='fasta'):
    module_coverage_frame = make_module_coverage_frame(annotations, module_nets, groupby_column)

    # make ETC frame
    etc_coverage_df = make_etc_coverage_df(etc_module_df, annotations, groupby_column)

    # make functional frame
    function_df = make_functional_df(annotations, function_heatmap_form, groupby_column)

    return module_coverage_frame, etc_coverage_df, function_df


def rename_genomes_to_taxa(function_df, labels, mag_order):
    function_df = function_df.copy()
    new_genome_column = [labels[i] for i in function_df['genome']]
    function_df['genome'] = pd.Series(new_genome_column, index=function_df.index)
    mag_order = [labels[i] for i in mag_order]
    return function_df, mag_order


def make_liquor_heatmap(module_coverage_frame, etc_coverage_df, function_df, mag_order=None, labels=None):
    module_coverage_heatmap = make_module_coverage_heatmap(module_coverage_frame, mag_order)
    etc_heatmap = make_etc_coverage_heatmap(etc_coverage_df, mag_order=mag_order)
    if labels is not None:
        function_df, mag_order = rename_genomes_to_taxa(function_df, labels, mag_order)
    function_heatmap = make_functional_heatmap(function_df, mag_order)

    liquor = alt.hconcat(alt.hconcat(module_coverage_heatmap, etc_heatmap), function_heatmap)
    return liquor


def make_liquor_df(module_coverage_frame, etc_coverage_df, function_df):
    liquor_df = pd.concat([module_coverage_frame.pivot(index='genome', columns='module_name', values='step_coverage'),
                           etc_coverage_df.pivot(index='genome', columns='complex_module_name',
                                                 values='percent_coverage'),
                           function_df.pivot(index='genome', columns='category_function_name', values='present')],
                          axis=1, sort=False)
    return liquor_df


def get_phylum_and_most_specific(taxa_str):
    taxa_ranks = [i[3:] for i in taxa_str.split(';')]
    phylum = taxa_ranks[1]
    most_specific_rank = TAXONOMY_LEVELS[sum([len(i) > 0 for i in taxa_ranks])-1]
    most_specific_taxa = taxa_ranks[sum([len(i) > 0 for i in taxa_ranks]) - 1]
    if most_specific_rank == 'd':
        return 'd__%s;p__' % most_specific_taxa
    if most_specific_rank == 'p':
        return 'p__%s;c__' % most_specific_taxa
    else:
        return 'p__%s;%s__%s' % (phylum, most_specific_rank, most_specific_taxa)


def make_strings_no_repeats(genome_taxa_dict):
    labels = dict()
    seen = Counter()
    for genome, taxa_string in genome_taxa_dict.items():
        final_taxa_string = '%s_%s' % (taxa_string, str(seen[taxa_string]))
        seen[taxa_string] += 1
        labels[genome] = final_taxa_string
    return labels


def summarize_genomes(input_file, trna_path=None, rrna_path=None, output_dir='.', groupby_column='fasta'):
    start_time = datetime.now()

    # read in data
    annotations = pd.read_csv(input_file, sep='\t', index_col=0)
    if 'bin_taxnomy' in annotations:
        annotations = annotations.sort_values('bin_taxonomy')

    if trna_path is None:
        trna_frame = None
    else:
        trna_frame = pd.read_csv(trna_path, sep='\t')
    if rrna_path is None:
        rrna_frame = None
    else:
        rrna_frame = pd.read_csv(rrna_path, sep='\t')

    # get db_locs and read in dbs
    db_locs = get_database_locs()
    if 'genome_summary_form' not in db_locs:
        raise ValueError('Genome summary form location must be set in order to summarize genomes')
    if 'module_step_form' not in db_locs:
        raise ValueError('Module step form location must be set in order to summarize genomes')
    if 'function_heatmap_form' not in db_locs:
        raise ValueError('Functional heat map form location must be set in order to summarize genomes')

    # read in dbs
    genome_summary_form = pd.read_csv(db_locs['genome_summary_form'], sep='\t')
    genome_summary_form = genome_summary_form.drop('potential_amg', axis=1)
    module_steps_form = pd.read_csv(db_locs['module_step_form'], sep='\t')
    function_heatmap_form = pd.read_csv(db_locs['function_heatmap_form'], sep='\t')
    etc_module_df = pd.read_csv(db_locs['etc_module_database'], sep='\t')
    print('%s: Retrieved database locations and descriptions' % (str(datetime.now() - start_time)))

    # make output folder
    mkdir(output_dir)

    # make genome stats
    genome_stats = make_genome_stats(annotations, rrna_frame, trna_frame)
    genome_stats.to_csv(path.join(output_dir, 'genome_stats.tsv'), sep='\t', index=None)
    print('%s: Calculated genome statistics' % (str(datetime.now() - start_time)))

    # make genome metabolism summary
    genome_summary = path.join(output_dir, 'metabolism_summary.xlsx')
    summarized_genomes = make_genome_summary(annotations, genome_summary_form, trna_frame, rrna_frame, groupby_column)
    write_summarized_genomes_to_xlsx(summarized_genomes, genome_summary)
    print('%s: Generated genome metabolism summary' % (str(datetime.now() - start_time)))

    # make liquor
    if 'bin_taxonomy' in annotations:
        genome_order = get_ordered_uniques(annotations.sort_values('bin_taxonomy')[groupby_column])
        # if gtdb format then get phylum and most specific
        if all([i[:3] == 'd__' and len(i.split(';')) == 7 for i in annotations['bin_taxonomy']]):
            taxa_str_parser = get_phylum_and_most_specific
        # else just throw in what is there
        else:
            taxa_str_parser = lambda x: x
        labels = make_strings_no_repeats({row[groupby_column]: taxa_str_parser(row['bin_taxonomy'])
                                          for _, row in annotations.iterrows()})
    else:
        genome_order = get_ordered_uniques(annotations.sort_values(groupby_column)[groupby_column])
        labels = None

    # make module coverage frame
    module_nets = {module: build_module_net(module_df)
                   for module, module_df in module_steps_form.groupby('module') if module in HEATMAP_MODULES}

    if len(genome_order) > GENOMES_PER_LIQUOR:
        module_coverage_dfs = list()
        etc_coverage_dfs = list()
        function_dfs = list()
        # generates slice start and slice end to grab from genomes and labels from 0 to end of genome order
        pairwise_iter = pairwise(list(range(0, len(genome_order), GENOMES_PER_LIQUOR)) + [len(genome_order) - 1])
        for i, (start, end) in enumerate(pairwise_iter):
            genomes = genome_order[start:end]
            annotations_subset = annotations.loc[[genome in genomes for genome in annotations[groupby_column]]]
            dfs = fill_liquor_dfs(annotations_subset, module_nets, etc_module_df, function_heatmap_form,
                                  groupby_column='fasta')
            module_coverage_df_subset, etc_coverage_df_subset, function_df_subset = dfs
            module_coverage_dfs.append(module_coverage_df_subset)
            etc_coverage_dfs.append(etc_coverage_df_subset)
            function_dfs.append(function_df_subset)
            liquor = make_liquor_heatmap(module_coverage_df_subset, etc_coverage_df_subset, function_df_subset,
                                         genomes, labels)
            liquor.save(path.join(output_dir, 'product_%s.html' % i))
        liquor_df = make_liquor_df(pd.concat(module_coverage_dfs), pd.concat(etc_coverage_dfs), pd.concat(function_dfs))
        liquor_df.to_csv(path.join(output_dir, 'product.tsv'), sep='\t')
    else:
        module_coverage_df, etc_coverage_df, function_df = fill_liquor_dfs(annotations, module_nets,
                                                                           etc_module_df,
                                                                           function_heatmap_form,
                                                                           groupby_column='fasta')
        liquor_df = make_liquor_df(module_coverage_df, etc_coverage_df, function_df)
        liquor_df.to_csv(path.join(output_dir, 'product.tsv'), sep='\t')
        liquor = make_liquor_heatmap(module_coverage_df, etc_coverage_df, function_df, genome_order, labels)
        liquor.save(path.join(output_dir, 'product.html'))
    print('%s: Generated product heatmap and table' % (str(datetime.now() - start_time)))
    print("%s: Completed distillation" % str(datetime.now() - start_time))



#!/usr/bin/env python3

import argparse

from mag_annotator.annotate_bins import annotate_bins_cmd, annotate_called_genes
from mag_annotator.summarize_genomes import summarize_genomes
from mag_annotator.pull_sequences import pull_sequences

# TODO: refactor parses to limit duplication

if __name__ == '__main__':
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    subparsers = parser.add_subparsers()
    annotate_parser = subparsers.add_parser('annotate', help="Annotate genomes/contigs/bins/MAGs",
                                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    annotate_genes_parser = subparsers.add_parser('annotate_genes', help="Annotate already called genes, limited "
                                                                         "functionality compared to annotate",
                                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    distill_parser = subparsers.add_parser('distill', help="Summarize metabolic content of annotated genomes",
                                           formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    strainer_parser = subparsers.add_parser('strainer', help="Strain annotations down to genes of interest",
                                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # parser for annotating mags, you know the real thing
    annotate_parser.add_argument('-i', '--input_fasta',
                                 help="fasta file, optionally with wildcards to point to multiple fastas",
                                 required=True)
    annotate_parser.add_argument('-o', '--output_dir', help="output directory")
    annotate_parser.add_argument('--min_contig_size', type=int, default=2500,
                                 help='minimum contig size to be used for gene prediction')
    annotate_parser.add_argument('--bit_score_threshold', type=int, default=60,
                                 help='minimum bitScore of search to retain hits')
    annotate_parser.add_argument('--rbh_bit_score_threshold', type=int, default=350,
                                 help='minimum bitScore of reverse best hits to retain hits')
    annotate_parser.add_argument('--custom_db_name', action='append', help="Names of custom databases, can be used"
                                                                           "multiple times.")
    annotate_parser.add_argument('--custom_fasta_loc', action='append',
                                 help="Location of fastas to annotated against, can be used multiple times but"
                                      "must match nubmer of custom_db_name's")
    annotate_parser.add_argument('--gtdb_taxonomy', action='append', default=[],
                                 help='Summary file from gtdbtk taxonomy assignment from bins, can be used multiple'
                                      'times')
    annotate_parser.add_argument('--checkm_quality', action='append', default=[],
                                 help='Summary of of checkM quality assessment from bins, can be used multiple times')
    annotate_parser.add_argument('--use_uniref', action='store_true', default=False,
                                 help='Annotate these fastas against UniRef, drastically decreases run time and memory '
                                      'requirements')
    annotate_parser.add_argument('--low_mem_mode', action='store_true', default=False,
                                 help='Skip annotating with uniref and use kofam instead of KEGG genes even if '
                                      'provided. Drastically decreases memory usage')
    annotate_parser.add_argument('--skip_trnascan', action='store_true', default=False)
    annotate_parser.add_argument('--keep_tmp_dir', action='store_true', default=False)
    annotate_parser.add_argument('--threads', type=int, default=10, help='number of processors to use')
    annotate_parser.add_argument('--verbose', action='store_true', default=False)
    annotate_parser.set_defaults(func=annotate_bins_cmd)

    # parser for annotating already called genes
    annotate_genes_parser.add_argument('-i', '--input_faa', help="fasta file, optionally with wildcards to point to"
                                                                 "individual MAGs", required=True)
    annotate_genes_parser.add_argument('-o', '--output_dir', help="output directory")
    annotate_genes_parser.add_argument('--bit_score_threshold', type=int, default=60,
                                       help='minimum bitScore of search to retain hits')
    annotate_genes_parser.add_argument('--rbh_bit_score_threshold', type=int, default=350,
                                       help='minimum bitScore of reverse best hits to retain hits')
    annotate_genes_parser.add_argument('--custom_db_name', action='append', help="Names of custom databases, can be "
                                                                                 "used multiple times.")
    annotate_genes_parser.add_argument('--custom_fasta_loc', action='append',
                                       help="Location of fastas to annotated against, can be used multiple times but"
                                            "must match nubmer of custom_db_name's")
    annotate_genes_parser.add_argument('--use_uniref', action='store_true', default=False,
                                       help='Annotate these fastas against UniRef, drastically decreases run time and '
                                            'memory requirements')
    annotate_genes_parser.add_argument('--low_mem_mode', action='store_true', default=False,
                                       help='Skip annotating with uniref and use kofam instead of KEGG genes even if '
                                            'provided. Drastically decreases memory usage')
    annotate_genes_parser.add_argument('--keep_tmp_dir', action='store_true', default=False)
    annotate_genes_parser.add_argument('--threads', type=int, default=10, help='number of processors to use')
    annotate_genes_parser.add_argument('--verbose', action='store_true', default=False)
    annotate_genes_parser.set_defaults(func=annotate_called_genes)

    # parser for summarizing genomes
    distill_parser.add_argument("-i", "--input_file", help="Annotations path")
    distill_parser.add_argument("-o", "--output_dir", help="Directory to write summarized genomes")
    distill_parser.add_argument("--rrna_path", help="rRNA output from annotation")
    distill_parser.add_argument("--trna_path", help="tRNA output from annotation")
    distill_parser.add_argument("--groupby_column", help="Column from annotations to group as organism units",
                                default='fasta')
    distill_parser.set_defaults(func=summarize_genomes)

    # parser for getting genes
    input_group = strainer_parser.add_argument_group('Input and output files')
    input_group.add_argument('-i', '--input_annotations', required=True, help='annotations file to pull genes from')
    input_group.add_argument('-f', '--input_fasta', required=True, help='fasta file to filter')
    input_group.add_argument('-o', '--output_fasta', default='pull_genes.fasta',
                             help='location to write filtered fasta')
    name_group = strainer_parser.add_argument_group('Specific names to keep')
    name_group.add_argument('--fastas', nargs='*', default=None, help='space separated list of fastas to keep')
    name_group.add_argument('--scaffolds', nargs='*', default=None,
                            help='space separated list of scaffolds to keep')
    name_group.add_argument('--genes', nargs='*', default=None, help='space separated list of genes to keep')
    annotation_group = strainer_parser.add_argument_group('Annotation filters')
    annotation_group.add_argument('--identifiers', nargs='*', default=None, help='database identifiers to keep')
    annotation_group.add_argument('--categories', nargs='*', default=None,
                                  help='distillate categories to keep genes from')
    dram_group = strainer_parser.add_argument_group('DRAM based filters')
    dram_group.add_argument('--taxonomy', nargs='*', default=None,
                            help='Level of GTDBTk taxonomy to keep (e.g. c__Clostridia), space separated list')
    dram_group.add_argument('--completeness', default=None, type=float,
                            help='Minimum completeness of genome to keep genes')
    dram_group.add_argument('--contamination', default=None, type=float,
                            help='Maximum contamination of genome to keep genes')
    strainer_parser.set_defaults(func=pull_sequences)

    args = parser.parse_args()
    args_dict = {i: j for i, j in vars(args).items() if i != 'func'}
    args.func(**args_dict)



import pytest

import pandas as pd
import altair as alt

from mag_annotator.summarize_vgfs import filter_to_amgs, get_strand_switches, get_ids_from_row, make_viral_distillate, \
    make_vgf_order, make_amg_count_column, make_viral_functional_df, make_viral_functional_heatmap


@pytest.fixture()
def annotations():
    return pd.DataFrame([['scaffold_1', 'K99999', 'VTF', 1],
                         ['scaffold_1', 'K12345', 'M', 4],
                         ['scaffold_1', 'K00001', 'MFTJ', 3],
                         ['scaffold_1', 'K11111', 'MKE', 2]],
                        index=['gene_1', 'gene_2', 'gene_3', 'gene_4'],
                        columns=['scaffold', 'kegg_id', 'amg_flags', 'auxiliary_score'])


@pytest.fixture()
def pamgs():
    return pd.DataFrame([['scaffold_1', 'K00001', 'MFTJ', 3],
                         ['scaffold_1', 'K11111', 'MKE', 2]],
                        index=['gene_3', 'gene_4'],
                        columns=['scaffold', 'kegg_id', 'amg_flags', 'auxiliary_score'])


def test_filter_to_amgs(annotations, pamgs):
    test_pamgs = filter_to_amgs(annotations, remove_transposons=False, max_aux=3)
    pd.testing.assert_frame_equal(test_pamgs, pamgs)


def test_get_strand_switches():
    assert get_strand_switches(['+', '-', '+']) == 2
    assert get_strand_switches(['+', '+', '+']) == 0
    assert get_strand_switches(['-']) == 0
    assert get_strand_switches(['+', '+', '-', '-']) == 1


def test_get_ids_from_row():
    id_set1 = get_ids_from_row(pd.Series({'kegg_id': 'K00001,K00003'}))
    assert id_set1 == {'K00001', 'K00003'}
    id_set2 = get_ids_from_row(pd.Series({'kegg_hit': 'Some text and then [EC:0.0.0.0]; also [EC:1.1.1.1]'}))
    assert id_set2 == {'EC:0.0.0.0', 'EC:1.1.1.1'}
    id_set3 = get_ids_from_row(pd.Series({'peptidase_family': 'ABC1;BCD2'}))
    assert id_set3 == {'ABC1', 'BCD2'}
    id_set4 = get_ids_from_row(pd.Series({'cazy_hits': 'GH4 some things;GT6 other things'}))
    assert id_set4 == {'GH4', 'GT6'}


@pytest.fixture()
def genome_summary_frame():
    return pd.DataFrame(pd.DataFrame([['description', 'module1', 'main', 'header1', 'subheader1'],
                                      ['description2', 'module2', 'main', 'header1', 'subheader1'],
                                      ['description3', 'module3', 'other', 'header1', 'subheader1']],
                                     index=['K00001', 'K12345', 'K00001'],
                                     columns=['gene_description', 'module', 'sheet', 'header', 'subheader']))


def test_make_viral_distillate(pamgs, genome_summary_frame):
    test_viral_distillate = make_viral_distillate(pamgs, genome_summary_frame)
    viral_distillate = pd.DataFrame([['gene_3', 'scaffold_1', 'K00001', 'description', 'main', 'header1', 'subheader1',
                                      'module1', 3, 'MFTJ'],
                                     ['gene_3', 'scaffold_1', 'K00001', 'description3', 'other', 'header1',
                                      'subheader1', 'module3', 3, 'MFTJ'],
                                     ['gene_4', 'scaffold_1', '', '', '', '', '', '', 2, 'MKE']],
                                    columns=['gene', 'scaffold', 'gene_id', 'gene_description', 'category', 'header',
                                             'subheader', 'module', 'auxiliary_score', 'amg_flags'])
    pd.testing.assert_frame_equal(test_viral_distillate, viral_distillate)


def test_make_vgf_order(pamgs):
    assert make_vgf_order(pamgs) == ['scaffold_1']


def test_make_amg_count_column(pamgs):
    assert type(make_amg_count_column(pamgs)) is alt.Chart


@pytest.fixture()
def viral_functional_df():
    return pd.DataFrame([['main', 'module1', 'gene_3', 'K00001', 'scaffold_1', True],
                         ['other', 'module3', 'gene_3', 'K00001', 'scaffold_1', True]],
                        columns=['Category', 'Function', 'AMG Genes', 'Genes Present', 'Contig Name',
                                 'Present in Contig'])


def test_make_viral_functional_df(pamgs, genome_summary_frame, viral_functional_df):
    test_viral_function_df = make_viral_functional_df(pamgs, genome_summary_frame, 'scaffold')
    pd.testing.assert_frame_equal(test_viral_function_df, viral_functional_df)


def test_make_viral_functional_heatmap(viral_functional_df):
    assert type(make_viral_functional_heatmap(viral_functional_df)) is alt.HConcatChart



from datetime import datetime
from os import path, mkdir
import re

import pandas as pd
import numpy as np
from skbio.io import read as read_sequence
from skbio.io import write as write_sequence

from mag_annotator.database_handler import DatabaseHandler
from mag_annotator.annotate_bins import filter_db_locs, annotate_fastas
from mag_annotator.utils import get_database_locs, get_ids_from_annotation

VMAG_DBS_TO_ANNOTATE = ('kegg', 'kofam', 'kofam_ko_list', 'uniref', 'peptidase', 'pfam', 'dbcan', 'viral', 'vogdb')
VIRSORTER_COLUMN_NAMES = ['gene_name', 'start_position', 'end_position', 'length', 'strandedness',
                          'viral_protein_cluster_hit', 'viral_protein_cluster_hit_score',
                          'viral_protein_cluster_hit_evalue', 'viral_protein_cluster_category', 'pfam_hit',
                          'pfam_hit_score', 'pfam_hit_evalue', 'name']
VIRSORTER_HALLMARK_GENE_CATEGORIES = {'0', '3'}
VIRSORTER_VIRAL_LIKE_GENE_CATEGORIES = {'1', '4'}
TRANSPOSON_PFAMS = {'PF01609', 'PF00872', 'PF01610', 'PF01527', 'PF02371', 'PF01710', 'PF01385', 'PF01548', 'PF01526',
                    'PF01797', 'PF02899', 'PF05717', 'PF07592', 'PF03050', 'PF04754', 'PF04986', 'PF03400'}
CELL_ENTRY_CAZYS = {'CBM50', 'GH102', 'GH103', 'GH104', 'GH108', 'GH18', 'GH19', 'GH22', 'GH23', 'GH24', 'GH25', 'GH73',
                    'PL9', 'CBM12', 'CBM14', 'CBM18', 'CBM19', 'AA15', 'AA10', 'GH32', 'GH58', 'PL16'}
VIRAL_PEPTIDASES_MEROPS = {'A02H', 'A02G', 'A02F', 'A02E', 'A02D', 'A02C', 'A02B', 'A02A', 'A03B', 'A03A', 'A11B',
                           'A11A', 'A22B', 'A22A', 'A33', 'C01B', 'C01A', 'C02B', 'C02A', 'C03H', 'C03G', 'C03F',
                           'C03E', 'C03D', 'C03C', 'C03B', 'C03A', 'C04', 'C05', 'C06', 'C07', 'C08', 'C09', 'C104',
                           'C105', 'C107', 'C108', 'C113', 'C14B', 'C14A', 'C16B', 'C16A', 'C18', 'C19', 'C21', 'C23',
                           'C24', 'C26', 'C27', 'C28', 'C30', 'C31', 'C32', 'C33', 'C36', 'C37', 'C39', 'C40', 'C42',
                           'C44', 'C46', 'C48', 'C49', 'C51', 'C53', 'C57', 'C59', 'C60B', 'C60A', 'C62', 'C63', 'C71',
                           'C74', 'C76', 'C85B', 'C85A', 'C87', 'C89', 'C97', 'C99', 'G02', 'I02', 'I04', 'I08', 'I24',
                           'I25C', 'I25B', 'I25A', 'I29', 'I32', 'I36', 'I43', 'I50B', 'I50A', 'I51', 'I63', 'I75',
                           'I87', 'I91', 'M03C', 'M03B', 'M03A', 'M10C', 'M10B', 'M10A', 'M12C', 'M12B', 'M12A', 'M13',
                           'M14C', 'M14D', 'M14B', 'M14A', 'M15D', 'M15C', 'M15B', 'M15A', 'M16C', 'M16B', 'M16A',
                           'M20E', 'M20A', 'M20D', 'M20F', 'M20B', 'M20C', 'M23B', 'M23A', 'M27', 'M38', 'M41', 'M42',
                           'M43B', 'M43A', 'M44', 'M48C', 'M48B', 'M48A', 'M56', 'M60', 'M67C', 'M67B', 'M67A', 'M78',
                           'M79', 'M86', 'N01', 'N02', 'N04', 'N05', 'N07', 'N08', 'N09', 'N10E', 'N10D', 'N10C',
                           'N10B', 'N11', 'S01F', 'S01E', 'S01D', 'S01C', 'S01B', 'S01A', 'S03', 'S06', 'S07', 'S08C',
                           'S08B', 'S08A', 'S09D', 'S09C', 'S09B', 'S09A', 'S11', 'S12', 'S14', 'S16', 'S21', 'S24',
                           'S26C', 'S26B', 'S26A', 'S28', 'S29', 'S30', 'S31', 'S32', 'S33', 'S49C', 'S49B', 'S49A',
                           'S50', 'S53', 'S54', 'S62', 'S65', 'S69', 'S73', 'S74', 'S75', 'S77', 'S78', 'S80', 'S81',
                           'T01B', 'T01A', 'T03', 'U32', 'U40'}


def remove_bad_chars_fasta(fasta):
    new_headers = list()
    new_seqs = list()
    for seq in read_sequence(fasta, format='fasta'):
        new_header = seq.metadata['id']
        new_header = new_header.replace(';', '_').replace('=', '_')
        if new_header in new_headers:
            raise ValueError('Replacement of ; or = with _ generated reduntant sequence headers, must be modified '
                             'manually')
        seq.metadata['id'] = new_header
        new_seqs.append(seq)
    return new_seqs


def remove_bad_chars_virsorter_affi_contigs(virsorter_in):
    new_lines = list()
    for line in open(virsorter_in).readlines():
        if line.startswith('>'):
            line = line.lstrip('>')
            split_line = line.split('|')
            split_line[0] = split_line[0].replace(';', '_').replace('=', '_')
            new_lines.append('>%s' % '|'.join(split_line))
        else:
            split_line = line.split('|')
            split_line[0] = split_line[0].replace(';', '_').replace('=', '_')
            new_lines.append('|'.join(split_line))
    return ''.join(new_lines)


def remove_bad_chars(input_fasta=None, input_virsorter_affi_contigs=None, output=None):
    if input_fasta is None and input_virsorter_affi_contigs is None:
        raise ValueError('Must provide either input fasta or virsorter affi contigs')
    elif input_fasta is not None:
        if ';' in output or '=' in output:
            raise ValueError('Cannot have = or ; in output fasta to work with DRAM-v, please change output fasta name')
        new_seqs = remove_bad_chars_fasta(input_fasta)
        write_sequence((seq for seq in new_seqs), format='fasta', into=output)
    else:
        virsorter_out = remove_bad_chars_virsorter_affi_contigs(input_virsorter_affi_contigs)
        open(output, 'w').write(virsorter_out)


def get_virsorter_hits(virsorter_affi_contigs):
    raw_file = open(virsorter_affi_contigs).read()

    virsorter_rows = list()
    for entry in raw_file[1:].split('\n>'):
        split_entry = entry.strip().split('\n')
        entry_data = split_entry[0].split('|')
        entry_name = entry_data[0]
        if '=' in entry_name or ';' in entry_name:
            raise ValueError('FASTA headers must not have = or ; before the first space (%s). To run DRAM-v you must '
                             'rerun VIRSorter with = and ; removed from the headers or run DRAM-v.py '
                             'remove_bad_characters and then rerun DRAM-v' % entry_name)
        entry_genes = split_entry[1:]
        entry_rows = [i.split('|') + [entry_name] for i in entry_genes]
        virsorter_rows += entry_rows
    virsorter_hits = pd.DataFrame(virsorter_rows, columns=VIRSORTER_COLUMN_NAMES).set_index('gene_name')
    virsorter_hits.index = [i.replace('(', '_').replace(')', '_') for i in virsorter_hits.index]
    virsorter_hits['name'] = [i.replace('(', '_').replace(')', '_') for i in virsorter_hits['name']]
    return virsorter_hits


def get_overlap(row1, row2):
    overlap_start = np.max((row1['start_position'], row2['start_position']))
    overlap_end = np.min((row1['end_position'], row2['end_position']))
    overlap_len = overlap_end - overlap_start
    if overlap_len <= 0:
        return 0, 0
    else:
        return overlap_len / (row1['end_position'] - row1['start_position']), \
               overlap_len / (row2['end_position'] - row2['start_position'])


def is_transposon(pfam_hits):
    if pd.isna(pfam_hits):
        return False
    else:
        pfams = {i[1:-1].split('.')[0] for i in re.findall(r'\[PF\d\d\d\d\d.\d*\]', pfam_hits)}
        return len(pfams & TRANSPOSON_PFAMS) > 0


def get_gene_order(dram_genes, virsorter_genes, min_overlap=.70):
    dram_genes['start_position'] = dram_genes['start_position'].astype(int)
    dram_genes['end_position'] = dram_genes['end_position'].astype(int)
    dram_genes = dram_genes.sort_values('start_position')
    dram_gene_number = 0

    virsorter_genes['start_position'] = virsorter_genes['start_position'].astype(int)
    virsorter_genes['end_position'] = virsorter_genes['end_position'].astype(int)
    virsorter_genes = virsorter_genes.sort_values('start_position')
    virsorter_gene_number = 0

    merged_genes_rows = list()
    while (dram_gene_number < dram_genes.shape[0]) and (virsorter_gene_number < virsorter_genes.shape[0]):
        dram_row = dram_genes.iloc[dram_gene_number]
        dram_gene = dram_genes.index[dram_gene_number]
        virsorter_row = virsorter_genes.iloc[virsorter_gene_number]
        virsorter_gene = virsorter_genes.index[virsorter_gene_number]
        virsorter_gene_category = virsorter_row['viral_protein_cluster_category']
        # end if we are at the end of either list
        forward_overlap, reverse_overlap = get_overlap(dram_row, virsorter_row)
        if (forward_overlap > min_overlap) and (reverse_overlap > min_overlap):
            merged_genes_rows.append((dram_gene, virsorter_gene, virsorter_gene_category))
            dram_gene_number += 1
            virsorter_gene_number += 1
        else:  # if not significantly overlap then determine first gene
            if dram_row['start_position'] < virsorter_row['start_position']:
                merged_genes_rows.append((dram_gene, None, None))
                dram_gene_number += 1
            elif dram_row['start_position'] > virsorter_row['start_position']:
                merged_genes_rows.append((None, virsorter_gene, virsorter_gene_category))
                virsorter_gene_number += 1
            else:  # if the same start positions then shorter gene is the first
                if dram_row['end_position'] < virsorter_row['end_position']:
                    merged_genes_rows.append((dram_gene, None, None))
                    dram_gene_number += 1
                elif dram_row['end_position'] > virsorter_row['end_position']:
                    merged_genes_rows.append((None, virsorter_gene,
                                              virsorter_gene_category))
                    virsorter_gene_number += 1
                else:  # genes have the same start and end position so this shouldn't happen
                    raise ValueError('This should never happen: %s, %s, %s, %s' % (dram_row.start_position,
                                                                                   dram_row.end_position,
                                                                                   virsorter_row.start_position,
                                                                                   virsorter_row.end_position))
    # clean up and add extras
    # if at end of both then just end
    if (dram_gene_number == dram_genes.shape[0]) and (virsorter_gene_number == virsorter_genes.shape[0]):
        pass
    # if not at end of dram annotations then add the rest
    elif dram_gene_number != dram_genes.shape[0]:
        for i in range(dram_gene_number, dram_genes.shape[0]):
            dram_gene = dram_genes.index[i]
            merged_genes_rows.append((dram_gene, None, None))
    # if not at end of virsorter genes then add the rest
    elif virsorter_gene_number != virsorter_genes.shape[0]:
        for i in range(virsorter_gene_number, virsorter_genes.shape[0]):
            virsorter_row = virsorter_genes.iloc[i]
            merged_genes_rows.append((None, virsorter_genes.index[i],
                                      virsorter_row['viral_protein_cluster_category']))
    # how can we be not at the end of either?
    else:
        return ValueError('How is this even possible?')
    return merged_genes_rows


def calculate_auxiliary_scores(gene_order):
    # right now saying that flanked means between here and the end of the contig
    gene_auxiliary_score_dict = dict()
    for i, (dram_gene, virsorter_gene, virsorter_category) in enumerate(gene_order):
        if dram_gene is not None:
            left_categories = [left_virsorter_category for left_dram_gene, left_viral_gene, left_virsorter_category
                               in gene_order[:i] if left_viral_gene is not None]
            hallmark_left = len(set(left_categories) & set(VIRSORTER_HALLMARK_GENE_CATEGORIES)) > 0
            viral_like_left = len(set(left_categories) & set(VIRSORTER_VIRAL_LIKE_GENE_CATEGORIES)) > 0

            right_categories = [right_virsorter_category for right_dram_gene, right_viral_gene, right_virsorter_category
                                in gene_order[i+1:] if right_viral_gene is not None]
            hallmark_right = len(set(right_categories) & set(VIRSORTER_HALLMARK_GENE_CATEGORIES)) > 0
            viral_like_right = len(set(right_categories) & set(VIRSORTER_VIRAL_LIKE_GENE_CATEGORIES)) > 0
            # if end of contig then 5:
            if i == 0 or i == (len(gene_order)-1):
                auxiliary_score = 5
            elif hallmark_left and hallmark_right:  # hallmark on both sides then cat 1
                auxiliary_score = 1
            # hallmark on one side and viral like on other then cat 2
            elif (hallmark_left and viral_like_right) or (viral_like_left and hallmark_right):
                auxiliary_score = 2
            # viral like on both side then cat 3
            elif viral_like_left and viral_like_right:
                auxiliary_score = 3
            # not end of contig and hallmark or viral like on other side
            elif (hallmark_left or viral_like_left) or (hallmark_right or viral_like_right):
                auxiliary_score = 4
            # if gene is the only virsorter viral like or hallmark on contig then make it a 4
            elif (virsorter_category in VIRSORTER_HALLMARK_GENE_CATEGORIES) or \
                 (virsorter_category in VIRSORTER_VIRAL_LIKE_GENE_CATEGORIES):
                auxiliary_score = 4
            else:  # if no viral like or hallmark genes then score is 5
                auxiliary_score = 5
            gene_auxiliary_score_dict[dram_gene] = auxiliary_score
    return gene_auxiliary_score_dict


def get_metabolic_flags(annotations, metabolic_genes, amgs, verified_amgs, scaffold_length_dict,
                        length_from_end=5000):
    flag_dict = dict()
    metabolic_genes = set(metabolic_genes)
    for scaffold, scaffold_annotations in annotations.groupby('scaffold'):
        # perc_xh = sum([i == 'Xh' if not pd.isna(i) else False for i in scaffold_annotations['vogdb_categories']]) \
        #           / scaffold_annotations.shape[0]
        # is_j = perc_xh >= 0.18
        for gene, row in scaffold_annotations.iterrows():
            # set up
            flags = ''
            gene_annotations = set(get_ids_from_annotation(pd.DataFrame(row).transpose()).keys())
            # is viral
            if not pd.isna(row['vogdb_categories']):
                if len({'Xr', 'Xs'} & set(row['vogdb_categories'].split(';'))) > 0:
                    flags += 'V'
            # is metabolic
            if len(metabolic_genes & gene_annotations) > 0:
                flags += 'M'
            # is this a reported AMG reported
            if len(gene_annotations & set(amgs)) > 0:
                if 'M' not in flags:
                    flags += 'M'
                flags += 'K'
            # is this a experimentally verified amg
            if len(gene_annotations & set(verified_amgs)) > 0:
                flags += 'E'
            # is this gene a normal viral cell host entry gene
            if len(gene_annotations & CELL_ENTRY_CAZYS) > 0:
                flags += 'A'
            # is gene a normal virus peptidase
            if len(gene_annotations & VIRAL_PEPTIDASES_MEROPS) > 0:
                flags += 'P'
            # if there is a transposon in the contig
            if scaffold_annotations['is_transposon'].any():
                flags += 'T'
            # within 5 kb of end of contig
            if (int(row['start_position']) < length_from_end) or \
               (int(row['end_position']) > (scaffold_length_dict[row['scaffold']] - length_from_end)):
                flags += 'F'
            # if is_j:
            #     flags += 'J'
            flag_dict[gene] = flags
        # get 3 metabolic genes in a row flag
        for i in range(len(scaffold_annotations)):  # this needs to be fixed. Will only give B to middle of 3 genes.
            if 0 < i < (len(scaffold_annotations) - 1):
                gene = scaffold_annotations.index[i]
                gene_flags = flag_dict[gene]
                previous_gene = scaffold_annotations.index[i - 1]
                previous_gene_flags = flag_dict[previous_gene]
                next_gene = scaffold_annotations.index[i + 1]
                next_gene_flags = flag_dict[next_gene]
                if 'M' in previous_gene_flags and 'M' in gene_flags and 'M' in next_gene_flags:
                    if 'B' not in flag_dict[previous_gene]:
                        flag_dict[previous_gene] += 'B'
                    if 'B' not in flag_dict[gene]:
                        flag_dict[gene] += 'B'
                    if 'B' not in flag_dict[next_gene]:
                        flag_dict[next_gene] += 'B'
    return flag_dict


def get_amg_ids(amg_frame):
    ko_amgs = {j.strip() for i in amg_frame['KO'].dropna() for j in i.strip().split(';')}
    ec_amgs = {j.strip() for i in amg_frame['EC'].dropna() for j in i.strip().split(';')}
    pfam_amgs = {j.strip() for i in amg_frame['PFAM'].dropna() for j in i.strip().split(';')}
    return ko_amgs | ec_amgs | pfam_amgs


def get_virsorter_affi_contigs_name(scaffold):
    prophage_match = re.search(r'_gene_\d*_gene_\d*-\d*-\d*-cat_[123456]$', scaffold)
    plain_match = re.search(r'-cat_[123456]$', scaffold)
    if prophage_match is not None:
        virsorter_scaffold_name = scaffold[:prophage_match.start()]
    elif plain_match is not None:
        virsorter_scaffold_name = scaffold[:plain_match.start()]
    else:
        raise ValueError("Can't find VIRSorter endings on fasta header: %s" % scaffold)
    return virsorter_scaffold_name


def annotate_vgfs(input_fasta, virsorter_affi_contigs=None, output_dir='.', min_contig_size=2500,
                  bit_score_threshold=60, rbh_bit_score_threshold=350, custom_db_name=(), custom_fasta_loc=(),
                  use_uniref=False, low_mem_mode=False, skip_trnascan=False, keep_tmp_dir=True,
                  threads=10, verbose=True):
    # set up
    start_time = datetime.now()
    print('%s: Viral annotation started' % str(datetime.now()))

    # get database locations
    db_locs = get_database_locs()
    db_handler = DatabaseHandler(db_locs['description_db'])
    db_locs_anno = filter_db_locs(db_locs, low_mem_mode, use_uniref, VMAG_DBS_TO_ANNOTATE)

    if virsorter_affi_contigs is not None:
        virsorter_hits = get_virsorter_hits(virsorter_affi_contigs)
    else:
        virsorter_hits = None

    # split sequences into seperate fastas
    mkdir(output_dir)
    contig_dir = path.join(output_dir, 'vMAGs')
    mkdir(contig_dir)
    contig_locs = list()
    for seq in read_sequence(input_fasta, format='fasta'):
        if len(seq) >= min_contig_size:
            if '=' in seq.metadata['id'] or ';' in seq.metadata['id']:
                raise ValueError('FASTA headers must not have = or ; before the first space (%s). To run DRAM-v you '
                                 'must rerun VIRSorter with = and ; removed from the headers or run DRAM-v.py '
                                 'remove_bad_characters and then rerun DRAM-v' % seq.metadata['id'])
            if virsorter_hits is not None:
                if get_virsorter_affi_contigs_name(seq.metadata['id']) not in virsorter_hits['name'].values:
                    raise ValueError("No virsorter calls found in %s for scaffold %s from input fasta" %
                                     (virsorter_affi_contigs, seq.metadata['id']))
            contig_loc = path.join(contig_dir, '%s.fasta' % seq.metadata['id'])
            write_sequence((i for i in [seq]), format='fasta', into=contig_loc)
            contig_locs.append(contig_loc)

    # annotate vMAGs
    rename_bins = False
    annotations = annotate_fastas(contig_locs, output_dir, db_locs_anno, db_handler, min_contig_size,
                                  bit_score_threshold, rbh_bit_score_threshold, custom_db_name, custom_fasta_loc,
                                  skip_trnascan, rename_bins, keep_tmp_dir, start_time, threads, verbose)
    print('%s: Annotations complete, processing annotations' % str(datetime.now() - start_time))

    # setting up scoring viral genes
    amg_database_frame = pd.read_csv(db_locs['amg_database'], sep='\t')
    genome_summary_form = pd.read_csv(db_locs['genome_summary_form'], sep='\t', index_col=0)
    genome_summary_form = genome_summary_form.loc[genome_summary_form.potential_amg]

    # add auxiliary score
    if virsorter_hits is not None:
        gene_virsorter_category_dict = dict()
        gene_auxiliary_score_dict = dict()
        for scaffold, dram_frame in annotations.groupby('scaffold'):
            virsorter_scaffold_name = get_virsorter_affi_contigs_name(scaffold)
            virsorter_frame = virsorter_hits.loc[virsorter_hits.name == virsorter_scaffold_name]
            gene_order = get_gene_order(dram_frame, virsorter_frame)
            gene_virsorter_category_dict.update({dram_gene: virsorter_category for dram_gene, _, virsorter_category in
                                                 gene_order if dram_gene is not None})
            gene_auxiliary_score_dict.update(calculate_auxiliary_scores(gene_order))
        annotations['virsorter_category'] = pd.Series(gene_virsorter_category_dict)
        annotations['auxiliary_score'] = pd.Series(gene_auxiliary_score_dict)

    # get metabolic flags
    scaffold_length_dict = {seq.metadata['id']: len(seq) for seq in read_sequence(input_fasta, format='fasta')}
    metabolic_genes = set(genome_summary_form.index)
    if 'pfam_hits' in annotations:
        annotations['is_transposon'] = [is_transposon(i) for i in annotations['pfam_hits']]
    else:
        annotations['is_transposon'] = False

    amgs = get_amg_ids(amg_database_frame)
    verified_amgs = get_amg_ids(amg_database_frame.loc[amg_database_frame.verified])
    annotations['amg_flags'] = pd.Series(get_metabolic_flags(annotations, metabolic_genes, amgs,
                                                             verified_amgs, scaffold_length_dict))

    # downgrade B flag auxiliary scores
    if virsorter_affi_contigs is not None:
        annotations['auxiliary_score'] = pd.Series({gene: (4 if 'B' in row['amg_flags'] and row['auxiliary_score'] < 4
                                                           else row['auxiliary_score'])
                                                    for gene, row in annotations.iterrows()})

    # write annotations
    annotations.to_csv(path.join(output_dir, 'annotations.tsv'), sep='\t')

    print("%s: Completed annotations" % str(datetime.now() - start_time))



from os import path, mkdir, remove
from shutil import copy2
from datetime import datetime
from shutil import move, rmtree
from glob import glob
import json
import gzip
import tarfile
import pandas as pd
from collections import defaultdict
from skbio import read as read_sequence
from skbio import write as write_sequence

from mag_annotator.utils import run_process, make_mmseqs_db, get_database_locs, download_file, merge_files, \
    get_config_loc, remove_prefix
from mag_annotator.database_handler import DatabaseHandler
from mag_annotator.database_setup import create_description_db


# TODO: check if dbcan or pfam is down, raise appropriate error
# TODO: upgrade to pigz?


def get_iso_date():
    return datetime.today().strftime('%Y%m%d')


def check_file_exists(db_loc):
    if db_loc is None:
        return True
    elif path.isfile(db_loc):
        return True
    else:
        raise ValueError("Database location does not exist: %s" % db_loc)


def make_header_dict_from_mmseqs_db(mmseqs_db):
    mmseqs_headers_handle = open('%s_h' % mmseqs_db, 'rb')
    mmseqs_headers = mmseqs_headers_handle.read().decode(errors='ignore')
    mmseqs_headers = [i.strip() for i in mmseqs_headers.strip().split('\n\x00') if len(i) > 0]
    return [{'id': i.split(' ')[0], 'description': i} for i in mmseqs_headers]


def generate_modified_kegg_fasta(kegg_fasta, gene_ko_link_loc=None):
    """Takes kegg fasta file and gene ko link file, adds kos not already in headers to headers"""
    genes_ko_dict = defaultdict(list)
    if gene_ko_link_loc is not None:
        if gene_ko_link_loc.endswith('.gz'):
            gene_ko_link_fh = gzip.open(gene_ko_link_loc, 'rt')
        else:
            gene_ko_link_fh = open(gene_ko_link_loc)
        for line in gene_ko_link_fh:
            gene, ko = line.strip().split()
            genes_ko_dict[gene].append(remove_prefix(ko, 'ko:'))
    for seq in read_sequence(kegg_fasta, format='fasta'):
        new_description = seq.metadata['description']
        for ko in genes_ko_dict[seq.metadata['id']]:
            if ko not in new_description:
                new_description += '; %s' % ko
        seq.metadata['description'] = new_description
        yield seq


def process_kegg_db(output_dir, kegg_loc, gene_ko_link_loc=None, download_date=None, threads=10, verbose=True):
    if download_date is None:
        download_date = get_iso_date()
    if gene_ko_link_loc is not None:
        # add KOs to end of header where KO is not already there
        kegg_mod_loc = path.join(output_dir, 'kegg.mod.fa')
        write_sequence(generate_modified_kegg_fasta(kegg_loc, gene_ko_link_loc),
                       format='fasta', into=kegg_mod_loc)
    else:
        kegg_mod_loc = kegg_loc
    # make mmseqsdb from modified kegg fasta
    kegg_mmseqs_db = path.join(output_dir, 'kegg.%s.mmsdb' % download_date)
    make_mmseqs_db(kegg_mod_loc, kegg_mmseqs_db, create_index=True, threads=threads, verbose=verbose)
    return kegg_mmseqs_db


def download_and_process_kofam_hmms(kofam_profile_tar_gz=None, output_dir='.', verbose=False):
    if kofam_profile_tar_gz is None:
        kofam_profile_tar_gz = path.join(output_dir, 'kofam_profiles.tar.gz')
        download_file('ftp://ftp.genome.jp/pub/db/kofam/profiles.tar.gz', kofam_profile_tar_gz, verbose=verbose)
    kofam_profiles = path.join(output_dir, 'kofam_profiles')
    mkdir(kofam_profiles)
    run_process(['tar', '-xzf', kofam_profile_tar_gz, '-C', kofam_profiles], verbose=verbose)
    merged_kofam_profiles = path.join(output_dir, 'kofam_profiles.hmm')
    merge_files(glob(path.join(kofam_profiles, 'profiles', '*.hmm')), merged_kofam_profiles)
    run_process(['hmmpress', '-f', merged_kofam_profiles], verbose=verbose)
    return merged_kofam_profiles


def download_and_process_kofam_ko_list(kofam_ko_list_gz=None, output_dir='.', verbose=False):
    if kofam_ko_list_gz is None:
        kofam_ko_list_gz = path.join(output_dir, 'kofam_ko_list.tsv.gz')
        download_file('ftp://ftp.genome.jp/pub/db/kofam/ko_list.gz', kofam_ko_list_gz, verbose=verbose)
    # TODO: fix this so that it is gunzipped to the path
    kofam_ko_list = path.join(output_dir, 'kofam_ko_list.tsv')
    run_process(['gunzip', kofam_ko_list_gz], verbose=verbose)
    return kofam_ko_list


def download_and_process_uniref(uniref_fasta_zipped=None, output_dir='.', uniref_version='90', threads=10,
                                verbose=True):
    """"""
    if uniref_fasta_zipped is None:  # download database if not provided
        uniref_fasta_zipped = path.join(output_dir, 'uniref%s.fasta.gz' % uniref_version)
        uniref_url = 'ftp://ftp.uniprot.org/pub/databases/uniprot/uniref/uniref%s/uniref%s.fasta.gz' % \
                     (uniref_version, uniref_version)
        download_file(uniref_url, uniref_fasta_zipped, verbose=verbose)
    uniref_mmseqs_db = path.join(output_dir, 'uniref%s.%s.mmsdb' % (uniref_version, get_iso_date()))
    make_mmseqs_db(uniref_fasta_zipped, uniref_mmseqs_db, create_index=True, threads=threads, verbose=verbose)
    return uniref_mmseqs_db


def process_mmspro(full_alignment, output_dir, db_name='db', threads=10, verbose=True):
    mmseqs_msa = path.join(output_dir, '%s.mmsmsa' % db_name)
    run_process(['mmseqs', 'convertmsa', full_alignment, mmseqs_msa], verbose=verbose)
    mmseqs_profile = path.join(output_dir, '%s.mmspro' % db_name)
    run_process(['mmseqs', 'msa2profile', mmseqs_msa, mmseqs_profile, '--match-mode', '1', '--threads', str(threads)],
                verbose=verbose)
    tmp_dir = path.join(output_dir, 'tmp')
    run_process(['mmseqs', 'createindex', mmseqs_profile, tmp_dir, '-k', '5', '-s', '7', '--threads', str(threads)],
                verbose=verbose)
    return mmseqs_profile


def download_and_process_pfam(pfam_full_zipped=None, output_dir='.', threads=10, verbose=True):
    if pfam_full_zipped is None:  # download database if not provided
        pfam_full_zipped = path.join(output_dir, 'Pfam-A.full.gz')
        download_file('ftp://ftp.ebi.ac.uk/pub/databases/Pfam/current_release/Pfam-A.full.gz', pfam_full_zipped,
                      verbose=verbose)
    pfam_profile = process_mmspro(pfam_full_zipped, output_dir, 'pfam', threads, verbose)
    return pfam_profile


def download_pfam_descriptions(output_dir='.', verbose=True):
    pfam_hmm_dat = path.join(output_dir, 'Pfam-A.hmm.dat.gz')
    download_file('ftp://ftp.ebi.ac.uk/pub/databases/Pfam/current_release/Pfam-A.hmm.dat.gz', pfam_hmm_dat,
                  verbose=verbose)
    return pfam_hmm_dat


def process_pfam_descriptions(pfam_hmm_dat):
    if pfam_hmm_dat.endswith('.gz'):
        f = gzip.open(pfam_hmm_dat, 'r').read().decode('utf-8')
    else:
        f = open(pfam_hmm_dat).read()
    entries = f.strip().split('//')
    description_list = list()
    for i, entry in enumerate(entries):
        if len(entry) > 0:
            entry = entry.split('\n')
            ascession = None
            description = None
            for line in entry:
                line = line.strip()
                if line.startswith('#=GF AC'):
                    ascession = line.split('   ')[-1]
                if line.startswith('#=GF DE'):
                    description = line.split('   ')[-1]
            description_list.append({'id': ascession, 'description': description})
    return description_list


def download_and_process_dbcan(dbcan_hmm=None, output_dir='.', dbcan_release='8', verbose=True):
    if dbcan_hmm is None:  # download database if not provided
        dbcan_hmm = path.join(output_dir, 'dbCAN-HMMdb-V%s.txt' % dbcan_release)
        download_file('http://bcb.unl.edu/dbCAN2/download/Databases/dbCAN-HMMdb-V%s.txt' % dbcan_release, dbcan_hmm,
                      verbose=verbose)
    run_process(['hmmpress', '-f', dbcan_hmm], verbose=verbose)
    return dbcan_hmm


def download_dbcan_descriptions(output_dir='.', upload_date='07312019', verbose=True):
    dbcan_fam_activities = path.join(output_dir, 'CAZyDB.%s.fam-activities.txt' % upload_date)
    download_file('http://bcb.unl.edu/dbCAN2/download/Databases/CAZyDB.%s.fam-activities.txt' % upload_date,
                  dbcan_fam_activities, verbose=verbose)
    return dbcan_fam_activities


def process_dbcan_descriptions(dbcan_fam_activities):
    f = open(dbcan_fam_activities)
    description_list = list()
    for line in f.readlines():
        if not line.startswith('#') and len(line.strip()) != 0:
            line = line.strip().split()
            if len(line) == 1:
                description = line[0]
            elif line[0] == line[1]:
                description = ' '.join(line[1:])
            else:
                description = ' '.join(line)
            description_list.append({'id': line[0], 'description': description.replace('\n', ' ')})
    return description_list


def download_and_process_viral_refseq(merged_viral_faas=None, output_dir='.', viral_files=2, threads=10, verbose=True):
    """Can only download newest version"""
    # download all of the viral protein files, need to know the number of files
    # TODO: Make it so that you don't need to know number of viral files in refseq viral

    if merged_viral_faas is None:  # download database if not provided
        faa_base_name = 'viral.%s.protein.faa.gz'
        viral_faa_glob = path.join(output_dir, faa_base_name % '*')
        for number in range(viral_files):
            number += 1
            refseq_url = 'ftp://ftp.ncbi.nlm.nih.gov/refseq/release/viral/viral.%s.protein.faa.gz' % number
            refseq_faa = path.join(output_dir, faa_base_name % number)
            download_file(refseq_url, refseq_faa, verbose=verbose)

        # then merge files from above
        merged_viral_faas = path.join(output_dir, 'viral.merged.protein.faa.gz')
        run_process(['cat %s > %s' % (' '.join(glob(viral_faa_glob)), merged_viral_faas)], shell=True)

    # make mmseqs database
    refseq_viral_mmseqs_db = path.join(output_dir, 'refseq_viral.%s.mmsdb' % get_iso_date())
    make_mmseqs_db(merged_viral_faas, refseq_viral_mmseqs_db, create_index=True, threads=threads, verbose=verbose)
    return refseq_viral_mmseqs_db


def download_and_process_merops_peptidases(peptidase_faa=None, output_dir='.', threads=10, verbose=True):
    if peptidase_faa is None:  # download database if not provided
        peptidase_faa = path.join(output_dir, 'merops_peptidases_nr.faa')
        merops_url = 'ftp://ftp.ebi.ac.uk/pub/databases/merops/current_release/pepunit.lib'
        download_file(merops_url, peptidase_faa, verbose=verbose)
    peptidase_mmseqs_db = path.join(output_dir, 'peptidases.%s.mmsdb' % get_iso_date())
    make_mmseqs_db(peptidase_faa, peptidase_mmseqs_db, create_index=True, threads=threads, verbose=verbose)
    return peptidase_mmseqs_db


def download_and_process_vogdb(vog_hmm_targz=None, output_dir='.', vogdb_release='latest', verbose=True):
    if vog_hmm_targz is None:
        vog_hmm_targz = path.join(output_dir, 'vog.hmm.tar.gz')
        vogdb_url = 'http://fileshare.csb.univie.ac.at/vog/%s/vog.hmm.tar.gz' % vogdb_release
        download_file(vogdb_url, vog_hmm_targz, verbose=verbose)
    hmm_dir = path.join(output_dir, 'vogdb_hmms')
    mkdir(hmm_dir)
    vogdb_targz = tarfile.open(vog_hmm_targz)
    vogdb_targz.extractall(hmm_dir)
    vog_hmms = path.join(output_dir, 'vog_%s_hmms.txt' % vogdb_release)
    merge_files(glob(path.join(hmm_dir, 'VOG*.hmm')), vog_hmms)
    run_process(['hmmpress', '-f', vog_hmms], verbose=verbose)
    return vog_hmms


def download_vog_annotations(output_dir, vogdb_version='latest', verbose=True):
    vog_annotations = path.join(output_dir, 'vog_annotations_%s.tsv.gz' % vogdb_version)
    download_file('http://fileshare.csb.univie.ac.at/vog/%s/vog.annotations.tsv.gz' % vogdb_version,
                  vog_annotations, verbose=verbose)
    return vog_annotations


def process_vogdb_descriptions(vog_annotations):
    annotations_table = pd.read_csv(vog_annotations, sep='\t', index_col=0)
    annotations_list = [{'id': vog, 'description': '%s; %s' % (row['ConsensusFunctionalDescription'],
                                                               row['FunctionalCategory'])}
                        for vog, row in annotations_table.iterrows()]
    return annotations_list


def download_and_process_genome_summary_form(output_dir, branch='master', verbose=True):
    genome_summary_form = path.join(output_dir, 'genome_summary_form.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/shafferm/DRAM/%s/data/genome_summary_form.tsv' % branch,
                  genome_summary_form, verbose=verbose)
    return genome_summary_form


def download_and_process_module_step_form(output_dir, branch='master', verbose=True):
    function_heatmap_form = path.join(output_dir, 'module_step_form.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/shafferm/DRAM/%s/data/module_step_form.tsv' % branch,
                  function_heatmap_form, verbose=verbose)
    return function_heatmap_form


def download_and_process_etc_module_database(output_dir, branch='master', verbose=True):
    etc_module_database = path.join(output_dir, 'etc_mdoule_database.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/shafferm/DRAM/%s/data/etc_module_database.tsv' % branch,
                  etc_module_database, verbose=verbose)
    return etc_module_database


def download_and_process_function_heatmap_form(output_dir, branch='master', verbose=True):
    function_heatmap_form = path.join(output_dir, 'function_heatmap_form.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/shafferm/DRAM/%s/data/function_heatmap_form.tsv' % branch,
                  function_heatmap_form, verbose=verbose)
    return function_heatmap_form


def download_and_process_amg_database(output_dir, branch='master', verbose=True):
    amg_database = path.join(output_dir, 'amg_database.%s.tsv' % get_iso_date())
    download_file('https://raw.githubusercontent.com/shafferm/DRAM/%s/data/amg_database.tsv' % branch,
                  amg_database, verbose=verbose)
    return amg_database


def check_exists_and_add_to_location_dict(loc, name, dict_to_update):
    if loc is not None:  # if location give and exists then add to dict, else raise ValueError
        if check_file_exists(loc):
            dict_to_update[name] = path.realpath(loc)
    else:  # if location not given and is not in dict then set to none, else leave previous value
        if name not in dict_to_update:
            dict_to_update[name] = None
    return dict_to_update


def add_to_description_db(loc, name, get_description_list, db_handler):
    if loc is not None:  # if location give and exists then add to dict, else raise ValueError
        description_list = get_description_list(loc)
        db_handler.add_descriptions_to_database(description_list, name, clear_table=True)


def set_database_paths(kegg_db_loc=None, kofam_hmm_loc=None, kofam_ko_list_loc=None, uniref_db_loc=None,
                       pfam_db_loc=None, pfam_hmm_dat=None, dbcan_db_loc=None, dbcan_fam_activities=None,
                       viral_db_loc=None, peptidase_db_loc=None, vogdb_db_loc=None, vog_annotations=None,
                       description_db_loc=None, genome_summary_form_loc=None, module_step_form_loc=None,
                       etc_module_database_loc=None, function_heatmap_form_loc=None, amg_database_loc=None,
                       start_time=None, config_loc=None, use_current_locs=True, update_description_db=False):
    if start_time is None:
        start_time = datetime.now()
    print('%s: Setting database paths' % str(datetime.now() - start_time))
    if use_current_locs:
        db_dict = get_database_locs()
    else:
        db_dict = {}

    db_dict = check_exists_and_add_to_location_dict(kegg_db_loc, 'kegg', db_dict)
    db_dict = check_exists_and_add_to_location_dict(kofam_hmm_loc, 'kofam', db_dict)
    db_dict = check_exists_and_add_to_location_dict(kofam_ko_list_loc, 'kofam_ko_list', db_dict)
    db_dict = check_exists_and_add_to_location_dict(uniref_db_loc, 'uniref', db_dict)
    db_dict = check_exists_and_add_to_location_dict(pfam_db_loc, 'pfam', db_dict)
    db_dict = check_exists_and_add_to_location_dict(pfam_hmm_dat, 'pfam_hmm_dat', db_dict)
    db_dict = check_exists_and_add_to_location_dict(dbcan_db_loc, 'dbcan', db_dict)
    db_dict = check_exists_and_add_to_location_dict(dbcan_fam_activities, 'dbcan_fam_activities', db_dict)
    db_dict = check_exists_and_add_to_location_dict(viral_db_loc, 'viral', db_dict)
    db_dict = check_exists_and_add_to_location_dict(peptidase_db_loc, 'peptidase', db_dict)
    db_dict = check_exists_and_add_to_location_dict(vogdb_db_loc, 'vogdb', db_dict)
    db_dict = check_exists_and_add_to_location_dict(vog_annotations, 'vog_annotations', db_dict)

    db_dict = check_exists_and_add_to_location_dict(genome_summary_form_loc, 'genome_summary_form', db_dict)
    db_dict = check_exists_and_add_to_location_dict(module_step_form_loc, 'module_step_form', db_dict)
    db_dict = check_exists_and_add_to_location_dict(etc_module_database_loc, 'etc_module_database', db_dict)
    db_dict = check_exists_and_add_to_location_dict(function_heatmap_form_loc, 'function_heatmap_form', db_dict)
    db_dict = check_exists_and_add_to_location_dict(amg_database_loc, 'amg_database', db_dict)

    if description_db_loc is not None:
        db_dict['description_db'] = description_db_loc
    elif 'description_db' not in db_dict:
        db_dict['description_db'] = None
    print('%s: Database locations added to CONFIG' % str(datetime.now() - start_time))

    if update_description_db:
        populate_description_db(db_dict, start_time)
        check_file_exists(description_db_loc)
        print('%s: Database descriptions updated' % str(datetime.now() - start_time))

    # change data paths
    if config_loc is None:
        config_loc = get_config_loc()
    with open(config_loc, 'w') as f:
        f.write(json.dumps(db_dict))
    print('%s: Database locations set' % str(datetime.now() - start_time))


def populate_description_db(db_dict=None, start_time=None):
    if start_time is None:
        start_time = datetime.now()
        print('%s: Populating description database' % str(datetime.now() - start_time))
    # setup
    if db_dict is None:
        db_dict = get_database_locs()
    if path.exists(db_dict['description_db']):
        remove(db_dict['description_db'])
    create_description_db(db_dict['description_db'])
    db_handler = DatabaseHandler(db_dict['description_db'])
    print('%s: Database connection established' % str(datetime.now() - start_time))

    # fill database
    add_to_description_db(db_dict['kegg'], 'kegg_description', make_header_dict_from_mmseqs_db,
                          db_handler)
    print('%s: KEGG descriptions added to description database' % str(datetime.now() - start_time))
    add_to_description_db(db_dict['uniref'], 'uniref_description', make_header_dict_from_mmseqs_db,
                          db_handler)
    print('%s: UniRef descriptions added to description database' % str(datetime.now() - start_time))
    add_to_description_db(db_dict['pfam_hmm_dat'], 'pfam_description', process_pfam_descriptions,
                          db_handler)
    print('%s: PFAM descriptions added to description database' % str(datetime.now() - start_time))
    add_to_description_db(db_dict['dbcan_fam_activities'], 'dbcan_description',
                          process_dbcan_descriptions, db_handler)
    print('%s: dbCAN descriptions added to description database' % str(datetime.now() - start_time))
    add_to_description_db(db_dict['viral'], 'viral_description', make_header_dict_from_mmseqs_db,
                          db_handler)
    print('%s: RefSeq viral descriptions added to description database' % str(datetime.now() - start_time))
    add_to_description_db(db_dict['peptidase'], 'peptidase_description',
                          make_header_dict_from_mmseqs_db, db_handler)
    print('%s: MEROPS descriptions added to description database' % str(datetime.now() - start_time))
    add_to_description_db(db_dict['vog_annotations'], 'vogdb_description', process_vogdb_descriptions,
                          db_handler)
    print('%s: VOGdb descriptions added to description database' % str(datetime.now() - start_time))
    print('%s: Description database populated' % str(datetime.now() - start_time))


def prepare_databases(output_dir, kegg_loc=None, gene_ko_link_loc=None, kofam_hmm_loc=None, kofam_ko_list_loc=None,
                      kegg_download_date=None, uniref_loc=None, uniref_version='90', pfam_loc=None, pfam_hmm_dat=None,
                      dbcan_loc=None, dbcan_version='8', dbcan_fam_activities=None, dbcan_date='07312019',
                      viral_loc=None, peptidase_loc=None, vogdb_loc=None, vogdb_version='latest', vog_annotations=None,
                      genome_summary_form_loc=None, module_step_form_loc=None, etc_module_database_loc=None,
                      function_heatmap_form_loc=None, amg_database_loc=None, skip_uniref=False,
                      keep_database_files=False, branch='master', threads=10, verbose=True):
    start_time = datetime.now()
    print('%s: Database preparation started' % str(datetime.now()))
    # check that all given files exist
    check_file_exists(kegg_loc)
    check_file_exists(gene_ko_link_loc)
    check_file_exists(kofam_hmm_loc)
    check_file_exists(kofam_ko_list_loc)
    check_file_exists(uniref_loc)
    check_file_exists(pfam_loc)
    check_file_exists(pfam_hmm_dat)
    check_file_exists(dbcan_loc)
    check_file_exists(dbcan_fam_activities)
    check_file_exists(vogdb_loc)
    check_file_exists(viral_loc)
    check_file_exists(peptidase_loc)
    check_file_exists(genome_summary_form_loc)
    check_file_exists(module_step_form_loc)
    check_file_exists(function_heatmap_form_loc)
    check_file_exists(amg_database_loc)

    # setup
    if not path.isdir(output_dir):
        mkdir(output_dir)
    temporary = path.join(output_dir, 'database_files')
    mkdir(temporary)

    # get databases
    output_dbs = dict()
    if kegg_loc is not None:
        output_dbs['kegg_db_loc'] = process_kegg_db(temporary, kegg_loc, gene_ko_link_loc, kegg_download_date, threads,
                                                    verbose)
        print('%s: KEGG database processed' % str(datetime.now() - start_time))
    if not skip_uniref:
        output_dbs['uniref_db_loc'] = download_and_process_uniref(uniref_loc, temporary, uniref_version=uniref_version,
                                                                  threads=threads, verbose=verbose)
        print('%s: UniRef database processed' % str(datetime.now() - start_time))
    output_dbs['pfam_db_loc'] = download_and_process_pfam(pfam_loc, temporary,
                                                          threads=threads, verbose=verbose)
    print('%s: PFAM database processed' % str(datetime.now() - start_time))
    output_dbs['dbcan_db_loc'] = download_and_process_dbcan(dbcan_loc, temporary, dbcan_release=dbcan_version,
                                                            verbose=verbose)
    print('%s: dbCAN database processed' % str(datetime.now() - start_time))
    output_dbs['viral_db_loc'] = download_and_process_viral_refseq(viral_loc, temporary, threads=threads,
                                                                   verbose=verbose)
    print('%s: RefSeq viral database processed' % str(datetime.now() - start_time))
    output_dbs['peptidase_db_loc'] = download_and_process_merops_peptidases(peptidase_loc, temporary, threads=threads,
                                                                            verbose=verbose)
    print('%s: MEROPS database processed' % str(datetime.now() - start_time))
    output_dbs['vogdb_db_loc'] = download_and_process_vogdb(vogdb_loc, temporary, vogdb_release=vogdb_version,
                                                            verbose=verbose)
    print('%s: VOGdb database processed' % str(datetime.now() - start_time))
    output_dbs['kofam_hmm_loc'] = download_and_process_kofam_hmms(kofam_hmm_loc, temporary, verbose=verbose)
    print('%s: KOfam database processed' % str(datetime.now() - start_time))
    output_dbs['kofam_ko_list_loc'] = download_and_process_kofam_ko_list(kofam_ko_list_loc, temporary, verbose=verbose)
    print('%s: KOfam ko list processed' % str(datetime.now() - start_time))

    # get pfam, dbcan and vogdb descriptions
    if pfam_hmm_dat is None:
        pfam_hmm_dat = download_pfam_descriptions(output_dir, verbose=verbose)
    output_dbs['pfam_hmm_dat'] = pfam_hmm_dat
    print('%s: PFAM hmm dat processed' % str(datetime.now() - start_time))
    if dbcan_fam_activities is None:
        dbcan_fam_activities = download_dbcan_descriptions(output_dir, dbcan_date, verbose=verbose)
    output_dbs['dbcan_fam_activities'] = dbcan_fam_activities
    print('%s: dbCAN fam activities processed' % str(datetime.now() - start_time))
    if vog_annotations is None:
        vog_annotations = download_vog_annotations(output_dir, vogdb_version, verbose=verbose)
    output_dbs['vog_annotations'] = vog_annotations
    print('%s: VOGdb annotations processed' % str(datetime.now() - start_time))

    # add genome summary form and function heatmap form
    if genome_summary_form_loc is None:
        output_dbs['genome_summary_form_loc'] = download_and_process_genome_summary_form(temporary, branch, verbose)
    else:
        output_dbs['genome_summary_form_loc'] = genome_summary_form_loc
    if module_step_form_loc is None:
        output_dbs['module_step_form_loc'] = download_and_process_module_step_form(temporary, branch, verbose)
    else:
        output_dbs['module_step_form_loc'] = module_step_form_loc
    if etc_module_database_loc is None:
        output_dbs['etc_module_database_loc'] = download_and_process_etc_module_database(temporary, branch, verbose)
    else:
        output_dbs['etc_module_database_loc'] = etc_module_database_loc
    if function_heatmap_form_loc is None:
        output_dbs['function_heatmap_form_loc'] = download_and_process_function_heatmap_form(temporary, branch, verbose)
    else:
        output_dbs['function_heatmap_form_loc'] = function_heatmap_form_loc
    if amg_database_loc is None:
        output_dbs['amg_database_loc'] = download_and_process_amg_database(temporary, branch, verbose)
    else:
        output_dbs['amg_database_loc'] = amg_database_loc
    print('%s: DRAM databases and forms downloaded' % str(datetime.now() - start_time))

    # move all files from temporary to output that will be kept
    for db_name, output_db in output_dbs.items():
        for db_file in glob('%s*' % output_db):
            move(db_file, path.join(output_dir, path.basename(db_file)))
        output_dbs[db_name] = path.join(output_dir, path.basename(output_db))
    print('%s: Files moved to final destination' % str(datetime.now() - start_time))

    output_dbs['description_db_loc'] = path.realpath(path.join(output_dir, 'description_db.sqlite'))

    set_database_paths(**output_dbs, use_current_locs=False, update_description_db=True, start_time=start_time)

    if not keep_database_files:
        rmtree(temporary)
    print('%s: Database preparation completed' % str(datetime.now() - start_time))


def print_database_locations(db_locs=None):
    if db_locs is None:
        db_locs = get_database_locs()

    print('KEGG db: %s' % db_locs.get('kegg'))
    print('KOfam db: %s' % db_locs.get('kofam'))
    print('KOfam KO list: %s' % db_locs.get('kofam_ko_list'))
    print('UniRef db: %s' % db_locs.get('uniref'))
    print('Pfam db: %s' % db_locs.get('pfam'))
    print('Pfam hmm dat: %s' % db_locs.get('pfam_hmm_dat'))
    print('dbCAN db: %s' % db_locs.get('dbcan'))
    print('dbCAN family activities: %s' % db_locs.get('dbcan_fam_activities'))
    print('RefSeq Viral db: %s' % db_locs.get('viral'))
    print('MEROPS peptidase db: %s' % db_locs.get('peptidase'))
    print('VOGDB db: %s' % db_locs.get('vogdb'))
    print('VOG annotations: %s' % db_locs.get('vog_annotations'))
    print('Description db: %s' % db_locs.get('description_db'))
    print('Genome summary form: %s' % db_locs.get('genome_summary_form'))
    print('Module step form: %s' % db_locs.get('module_step_form'))
    print('ETC module database: %s' % db_locs.get('etc_module_database'))
    print('Function heatmap form: %s' % db_locs.get('function_heatmap_form'))
    print('AMG database: %s' % db_locs.get('amg_database'))


def update_dram_forms(output_dir, branch='master'):
    if not path.isdir(output_dir):
        mkdir(output_dir)

    form_locs = dict()
    form_locs['genome_summary_form_loc'] = download_and_process_genome_summary_form(output_dir, branch)
    form_locs['module_step_form_loc'] = download_and_process_module_step_form(output_dir, branch)
    form_locs['etc_module_database_loc'] = download_and_process_etc_module_database(output_dir, branch)
    form_locs['function_heatmap_form_loc'] = download_and_process_function_heatmap_form(output_dir, branch)
    form_locs['amg_database_loc'] = download_and_process_amg_database(output_dir, branch)
    set_database_paths(**form_locs, update_description_db=False)


def export_config(output_file=None):
    if output_file is None:
        print(open(get_config_loc()).read())
    else:
        copy2(get_config_loc(), output_file)


def import_config(config_loc):
    copy2(config_loc, get_config_loc())



import pytest

import os
import json

from mag_annotator.utils import get_database_locs
from mag_annotator.database_processing import check_file_exists, make_header_dict_from_mmseqs_db, \
    generate_modified_kegg_fasta, process_kegg_db, process_mmspro, process_pfam_descriptions, \
    process_dbcan_descriptions, process_vogdb_descriptions, check_exists_and_add_to_location_dict, export_config, \
    print_database_locations, set_database_paths


def test_check_file_exists():
    assert check_file_exists(None)
    assert check_file_exists(os.path.join('tests', 'data', 'fake_gff.gff'))
    with pytest.raises(ValueError):
        check_file_exists(os.path.join('tests', 'data', 'a_fake_nonexistent_file.txt'))


def test_make_header_dict_from_mmseqs_db():
    test_headers = make_header_dict_from_mmseqs_db(os.path.join('tests', 'data', 'NC_001422.mmsdb'))
    assert test_headers == [{'id': 'NP_040705.1', 'description': 'NP_040705.1 B [Escherichia virus phiX174]'},
                            {'id': 'NP_040704.1', 'description': 'NP_040704.1 A* [Escherichia virus phiX174]'},
                            {'id': 'NP_040703.1', 'description': 'NP_040703.1 A [Escherichia virus phiX174]'},
                            {'id': 'NP_040713.1', 'description': 'NP_040713.1 H [Escherichia virus phiX174]'},
                            {'id': 'NP_040712.1', 'description': 'NP_040712.1 G [Escherichia virus phiX174]'},
                            {'id': 'NP_040711.1', 'description': 'NP_040711.1 F [Escherichia virus phiX174]'},
                            {'id': 'NP_040710.1', 'description': 'NP_040710.1 J [Escherichia virus phiX174]'},
                            {'id': 'NP_040709.1', 'description': 'NP_040709.1 E [Escherichia virus phiX174]'},
                            {'id': 'NP_040708.1', 'description': 'NP_040708.1 D [Escherichia virus phiX174]'},
                            {'id': 'NP_040707.1', 'description': 'NP_040707.1 C [Escherichia virus phiX174]'},
                            {'id': 'NP_040706.1', 'description': 'NP_040706.1 K [Escherichia virus phiX174]'}]


@pytest.fixture()
def phix_proteins():
    return os.path.join('tests', 'data', 'NC_001422.faa')


@pytest.fixture()
def fake_gene_link_loc():
    return os.path.join('tests', 'data', 'fake_gene_ko.link')


def test_generate_modified_kegg_fasta(phix_proteins, fake_gene_link_loc):
    seqs = generate_modified_kegg_fasta(phix_proteins,
                                        fake_gene_link_loc)
    seqs_dict = {seq.metadata['id']: seq for seq in seqs}
    assert 'K14021' in seqs_dict['NP_040705.1'].metadata['description']
    assert 'K04464' in seqs_dict['NP_040708.1'].metadata['description']


def test_process_kegg_db(tmpdir, phix_proteins, fake_gene_link_loc):
    processed_kegg_dir = tmpdir.mkdir('process_kegg_test')
    kegg_db = process_kegg_db(processed_kegg_dir, phix_proteins, fake_gene_link_loc, download_date='today')
    assert os.path.isfile(kegg_db)
    assert os.path.isfile('%s_h' % kegg_db)


def test_process_mmspro(tmpdir):
    processed_mmspro = tmpdir.mkdir('process_mmspro')
    process_mmspro(os.path.join('tests', 'data', 'Pfam-A_subset.full.gz'), processed_mmspro, 'fake', 1, False)
    assert os.path.isfile(os.path.join(processed_mmspro, 'fake.mmspro'))
    assert os.path.isfile(os.path.join(processed_mmspro, 'fake.mmspro_h'))
    assert os.path.isfile(os.path.join(processed_mmspro, 'fake.mmspro.idx'))
    assert os.path.isfile(os.path.join(processed_mmspro, 'fake.mmspro_h.index'))


def test_process_pfam_descriptions():
    description_list = process_pfam_descriptions(os.path.join('tests', 'data', 'Pfam-A_subset.hmm.dat.gz'))
    assert description_list == [{'id': 'PF10417.9', 'description': 'C-terminal domain of 1-Cys peroxiredoxin'},
                                {'id': 'PF12574.8', 'description': '120 KDa Rickettsia surface antigen'},
                                {'id': 'PF09847.9', 'description': 'Membrane protein of 12 TMs'},
                                {'id': 'PF00244.20', 'description': '14-3-3 protein'},
                                {'id': 'PF16998.5', 'description': '17 kDa outer membrane surface antigen'}]


def test_process_dbcan_descriptions():
    description_list = process_dbcan_descriptions(os.path.join('tests', 'data', 'CAZyDB.07312019.fam-activities.subset.txt'))
    assert description_list == [{'id': 'AA0', 'description': 'AA0'},
                                {'id': 'AA10', 'description': 'AA10 (formerly CBM33) proteins are copper-dependent '
                                                              'lytic polysaccharide monooxygenases (LPMOs); some '
                                                              'proteins have been shown to act on chitin, others on '
                                                              'cellulose; lytic cellulose monooxygenase '
                                                              '(C1-hydroxylating) (EC 1.14.99.54); lytic cellulose '
                                                              'monooxygenase (C4-dehydrogenating)(EC 1.14.99.56); '
                                                              'lytic chitin monooxygenase (EC 1.14.99.53)'},
                                {'id': 'AA11', 'description': 'AA11 proteins are copper-dependent lytic polysaccharide '
                                                              'monooxygenases (LPMOs); cleavage of chitin chains with '
                                                              'oxidation of C-1 has been demonstrated for a AA11 LPMO '
                                                              'from Aspergillus oryzae;'},
                                {'id': 'AA12', 'description': 'AA12 The pyrroloquinoline quinone-dependent oxidoreductase '
                                                              'activity was demonstrated for the CC1G_09525 protein of '
                                                              'Coprinopsis cinerea.'}]


def test_process_vogdb_descriptions():
    description_list = process_vogdb_descriptions(os.path.join('tests', 'data', 'vog_annotations_latest.subset.tsv.gz'))
    assert description_list == [{'id': 'VOG00001', 'description': 'sp|Q5UQJ2|YR863_MIMIV Putative ankyrin repeat '
                                                                  'protein R863; Xh'},
                                {'id': 'VOG00002', 'description': 'sp|Q9J4Z6|V244_FOWPN Putative ankyrin repeat '
                                                                  'protein FPV244; Xh'},
                                {'id': 'VOG00003', 'description': 'sp|Q91FD6|388R_IIV6 Putative MSV199 '
                                                                  'domain-containing protein 388R; Xu'},
                                {'id': 'VOG00004', 'description': 'sp|P51704|RPC1_BPHC1 Repressor protein CI; Xu'},
                                {'id': 'VOG00005', 'description': 'sp|P17766|POLG_PPVNA Genome polyprotein; XhXrXs'}]


def test_check_exists_and_add_to_location_dict(phix_proteins):
    test_dict1 = check_exists_and_add_to_location_dict(phix_proteins, 'fake', {})
    assert test_dict1 == {'fake': os.path.realpath(phix_proteins)}
    test_dict2 = check_exists_and_add_to_location_dict(None, 'fake', {})
    assert test_dict2 == {'fake': None}
    test_dict3 = check_exists_and_add_to_location_dict(None, 'fake', {'fake': 'abc123'})
    assert test_dict3 == {'fake': 'abc123'}


def test_export_config(capsys, tmpdir):
    export_config()
    out, err = capsys.readouterr()
    assert len(err) == 0
    assert 'description_db' in out
    assert out.startswith('{')
    assert out.strip().endswith('}')

    test_exported_config = os.path.join(tmpdir.mkdir('test_export'), 'test_CONFIG')
    export_config(test_exported_config)
    assert type(json.loads(open(test_exported_config).read())) is dict


def test_print_database_locations(capsys):
    print_database_locations()
    out, err = capsys.readouterr()
    assert 'Description db: ' in out
    assert len(err) == 0


def test_set_database_paths(tmpdir):
    test_config_dir = tmpdir.mkdir('test_config')
    # first test that adding nothing doesn't change CONFIG
    test_config = os.path.join(test_config_dir, 'CONFIG')
    pretest_db_dict = get_database_locs()
    set_database_paths(config_loc=test_config)
    test_db_dict = get_database_locs(test_config)
    assert type(test_db_dict) is dict
    assert pretest_db_dict == test_db_dict
    # test that adding something that doesn't exist throws error
    test_fake_database = os.path.join(test_config_dir, 'fake_database.mmsdb')
    with pytest.raises(ValueError):
        set_database_paths(kegg_db_loc=test_fake_database)
    # test that adding something real is really added
    kegg_loc = os.path.join('tests', 'data', 'fake_gff.gff')
    set_database_paths(kegg_db_loc=kegg_loc, config_loc=test_config)
    test_db_dict = get_database_locs(test_config)
    assert test_db_dict['kegg'] == os.path.realpath(kegg_loc)
    # test that adding something with use_current_locs False works
    set_database_paths(kegg_db_loc=kegg_loc, config_loc=test_config, use_current_locs=False)
    test_db_dict = get_database_locs(test_config)
    assert test_db_dict['kegg'] == os.path.realpath(kegg_loc)
    assert test_db_dict['description_db'] is None



from sqlalchemy import Column, String, create_engine
from sqlalchemy.ext.declarative import declarative_base

# TODO: Do all processing of descriptions here
# TODO: set up init statements that can parse the line into desired parameters

Base = declarative_base()

KEGG_DESCRIPTION_TABLE_NAME = 'kegg_description'


class KeggDescription(Base):
    __tablename__ = KEGG_DESCRIPTION_TABLE_NAME

    id = Column(String(20), primary_key=True, nullable=False, index=True)

    description = Column(String(100000))

    @property
    def serialize(self):
        return {
            'kegg_id': self.id,
            'kegg_description': self.description,
        }


UNIREF_DESCRIPTION_TABLE_NAME = 'uniref_description'


class UniRefDescription(Base):
    __tablename__ = UNIREF_DESCRIPTION_TABLE_NAME

    id = Column(String(40), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'kegg_id': self.id,
            'kegg_description': self.description,
        }


PFAM_DESCRIPTION_TABLE_NAME = 'pfam_description'


class PfamDescription(Base):
    __tablename__ = PFAM_DESCRIPTION_TABLE_NAME

    id = Column(String(12), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'pfam_id': self.id,
            'pfam_description': self.description,
        }


DBCAN_DESCRIPTION_TABLE_NAME = 'dbcan_description'


class DbcanDescription(Base):
    __tablename__ = DBCAN_DESCRIPTION_TABLE_NAME

    id = Column(String(30), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'dbcan_id': self.id,
            'dbcan_description': self.description,
        }


VIRAL_DESCRIPTION_TABLE_NAME = 'viral_description'


class ViralDescription(Base):
    __tablename__ = VIRAL_DESCRIPTION_TABLE_NAME

    id = Column(String(14), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'viral_id': self.id,
            'viral_description': self.description,
        }


PEPTIDASE_DESCRIPTION_TABLE_NAME = 'peptidase_description'


class PeptidaseDescription(Base):
    __tablename__ = PEPTIDASE_DESCRIPTION_TABLE_NAME

    id = Column(String(10), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'peptidase_id': self.id,
            'peptidase_description': self.description,
        }


VOGDB_DESCRIPTION_TABLE_NAME = 'vogdb_description'


class VOGDBDescription(Base):
    __tablename__ = VOGDB_DESCRIPTION_TABLE_NAME

    id = Column(String(10), primary_key=True, nullable=False, index=True)

    description = Column(String(1000))

    @property
    def serialize(self):
        return {
            'vogdb_id': self.id,
            'vogdb_description': self.description,
        }


def create_description_db(db_loc):
    engine = create_engine('sqlite:///%s' % db_loc)
    Base.metadata.create_all(engine)


TABLE_NAME_TO_CLASS_DICT = {KEGG_DESCRIPTION_TABLE_NAME: KeggDescription,
                            UNIREF_DESCRIPTION_TABLE_NAME: UniRefDescription,
                            PFAM_DESCRIPTION_TABLE_NAME: PfamDescription,
                            DBCAN_DESCRIPTION_TABLE_NAME: DbcanDescription,
                            VIRAL_DESCRIPTION_TABLE_NAME: ViralDescription,
                            PEPTIDASE_DESCRIPTION_TABLE_NAME: PeptidaseDescription,
                            VOGDB_DESCRIPTION_TABLE_NAME: VOGDBDescription}



import pytest

import os
import pandas as pd
import networkx as nx
import altair as alt

from mag_annotator.summarize_genomes import get_ordered_uniques, fill_genome_summary_frame, summarize_rrnas, \
    summarize_trnas, build_module_net, get_module_step_coverage, make_module_coverage_df, make_module_coverage_frame, \
    make_module_coverage_heatmap, pairwise, first_open_paren_is_all, split_into_steps, is_ko, make_module_network, \
    get_module_coverage, make_etc_coverage_df, make_etc_coverage_heatmap, make_functional_df, make_functional_heatmap, \
    fill_liquor_dfs, make_liquor_heatmap, make_liquor_df, make_genome_summary, write_summarized_genomes_to_xlsx,\
    get_phylum_and_most_specific


def test_get_ordered_uniques():
    assert get_ordered_uniques([1, 2, 3]) == [1, 2, 3]
    assert get_ordered_uniques([1, 1, 2, 3]) == [1, 2, 3]
    assert get_ordered_uniques([1, 2, 1, 3]) == [1, 2, 3]


@pytest.fixture()
def annotations():
    return pd.DataFrame([['genome', 'K00001'],
                         ['genome', pd.np.NaN]],
                        index=['genome_scaffold_1_1', 'genome_scaffold_1_2'],
                        columns=['fasta', 'kegg_id'])


@pytest.fixture()
def genome_summary_frame():
    return pd.DataFrame(pd.DataFrame([['K00001', 'description', 'module1', 'main', 'header1', 'subheader1'],
                                      ['K12345', 'description2', 'module2', 'main', 'header1', 'subheader1']],
                                     columns=['gene_id', 'gene_description', 'module', 'sheet', 'header', 'subheader']))


@pytest.fixture()
def summarized_genomes():
    return pd.DataFrame([['K00001', 'description', 'module1', 'main', 'header1', 'subheader1', 1],
                         ['K12345', 'description2', 'module2', 'main', 'header1', 'subheader1', 0]],
                        columns=['gene_id', 'gene_description', 'module', 'sheet', 'header', 'subheader', 'genome'])


def test_fill_genome_summary_frame(annotations, genome_summary_frame, summarized_genomes):
    test_frame = fill_genome_summary_frame(annotations, genome_summary_frame, 'fasta')
    pd.testing.assert_frame_equal(test_frame, summarized_genomes)


def test_summarize_rrnas():
    fake_rrnas = pd.DataFrame([['fake_NC_001422.1', 990, 1000, 101.0, '+', '16S rRNA', pd.np.NaN]],
                              columns=['scaffold', 'begin', 'end', 'e-value', 'strand', 'type', 'note'])
    test_rrna_frame = summarize_rrnas(fake_rrnas, 'scaffold')
    print(test_rrna_frame)
    real_rrna_frame = pd.DataFrame([['5S rRNA', '5S ribosomal RNA gene', 'rRNA', 'rRNA', '', '', 0],
                                    ['16S rRNA', '16S ribosomal RNA gene', 'rRNA', 'rRNA', '', '', 1],
                                    ['23S rRNA', '23S ribosomal RNA gene', 'rRNA', 'rRNA', '', '', 0]],
                                   columns=['gene_id', 'gene_description', 'module', 'sheet', 'header', 'subheader',
                                            'fake_NC_001422.1'])
    pd.testing.assert_frame_equal(test_rrna_frame, real_rrna_frame)


@pytest.fixture()
def fake_processed_trnas():
    return pd.read_csv(os.path.join('tests', 'data', 'fake_trnas.tsv'), sep='\t')


def test_summarize_trnas(fake_processed_trnas):
    test_summarized_trnas = summarize_trnas(fake_processed_trnas)
    summarized_trnas = pd.DataFrame([['Gly (GCC)', 'Gly tRNA with GCC Codon', 'Gly tRNA', 'tRNA', 'tRNA', '', 1., 0.],
                                     ['Gly (TCC)', 'Gly tRNA with TCC Codon', 'Gly tRNA', 'tRNA', 'tRNA', '', 0., 1.],
                                     ['Leu (TAA)', 'Leu tRNA with TAA Codon', 'Leu tRNA', 'tRNA', 'tRNA', '', 1., 0.],
                                     ['Tyr, pseudo (GTA)', 'Tyr pseudo tRNA with GTA Codon', 'Tyr tRNA', 'tRNA', 'tRNA',
                                      '', 0, 1]],
                                    columns=['gene_id', 'gene_description', 'module', 'sheet', 'header', 'subheader',
                                             'Cytophaga_hutchinsonii_ATCC_33406', 'other_organismii'])
    pd.testing.assert_frame_equal(test_summarized_trnas, summarized_trnas)


def test_make_genome_summary(annotations, genome_summary_frame, summarized_genomes):
    test_genome_summary = make_genome_summary(annotations, genome_summary_frame)
    assert type(test_genome_summary) is pd.DataFrame
    pd.testing.assert_frame_equal(test_genome_summary, summarized_genomes)


def test_write_summarized_genomes_to_xlsx(summarized_genomes, tmpdir):
    output_file = os.path.join(tmpdir.mkdir('genome_summary_xlsx'), 'genome_summary.xlsx')
    write_summarized_genomes_to_xlsx(summarized_genomes, output_file)
    assert os.path.isfile(output_file)


@pytest.fixture()
def test_module_net():
    module_frame = pd.DataFrame([['0,0', 'M12345', 'a module name', 'K00001'],
                                 ['1,0', 'M12345', 'a module name', 'K00002'],
                                 ['2,0', 'M12345', 'a module name', 'K00003']],
                                columns=['path', 'module', 'module_name', 'ko'])
    test_module_net = build_module_net(module_frame)
    return test_module_net


def test_build_module_net(test_module_net):
    real_module_net = nx.DiGraph(num_steps=0, module_id='M12345', module_name='a module name')
    real_module_net.add_edges_from([('0,0', 'end_step_0'), ('end_step_0', '1,0'),
                                    ('1,0', 'end_step_1'), ('end_step_1', '2,0'),
                                    ('2,0', 'end_step_2')])
    assert nx.is_isomorphic(test_module_net, real_module_net)


def test_get_module_step_coverage(test_module_net):
    test_coverages1 = get_module_step_coverage(set([]), test_module_net)
    assert test_coverages1 == (3, 0, 0, [])
    test_coverages2 = get_module_step_coverage({'K00001', 'K00003'}, test_module_net)
    assert test_coverages2 == (3, 2, 2/3, ['K00001', 'K00003'])
    test_coverages2 = get_module_step_coverage({'K00001', 'K00003', 'K00002', 'K12345'}, test_module_net)
    assert test_coverages2 == (3, 3, 1, ['K00001', 'K00002', 'K00003'])


@pytest.fixture()
def test_annotations_df():
    return pd.DataFrame([['', 'scaffold_1'],
                         ['K12345', 'scaffold_1'],
                         ['K00001', 'scaffold_1']],
                        index=['gene_1', 'gene_2', 'gene_3'], columns=['kegg_id', 'scaffold'])


def test_make_module_coverage_df(test_annotations_df, test_module_net):
    test_module_coverage_df = make_module_coverage_df(test_annotations_df, {'M12345': test_module_net})
    module_coverage_df = pd.DataFrame([['a module name', 3, 1, 1/3, 1, 'K00001', 'gene_3']],
                                      index=['M12345'],
                                      columns=['module_name', 'steps', 'steps_present', 'step_coverage', 'ko_count',
                                               'kos_present', 'genes_present'])
    pd.testing.assert_frame_equal(test_module_coverage_df, module_coverage_df)


@pytest.fixture()
def module_coverage_frame():
    return pd.DataFrame([['scaffold_1', 'M12345', 'a module name', 3, 1, 1 / 3, 1, 'K00001', 'gene_3']],
                        columns=['genome', 'module', 'module_name', 'steps', 'steps_present', 'step_coverage',
                                 'ko_count', 'kos_present', 'genes_present'])


def test_make_module_coverage_frame(test_annotations_df, test_module_net, module_coverage_frame):
    test_module_coverage_frame = make_module_coverage_frame(test_annotations_df, {'M12345': test_module_net},
                                                            groupby_column='scaffold')
    pd.testing.assert_frame_equal(test_module_coverage_frame, module_coverage_frame)


def test_make_module_coverage_heatmap(module_coverage_frame):
    hm = make_module_coverage_heatmap(module_coverage_frame)
    assert type(hm) is alt.Chart


def test_pairwise():
    assert list(pairwise([1, 2, 3])) == [(1, 2), (2, 3)]


def test_first_open_paren_is_all():
    assert first_open_paren_is_all('()')
    assert first_open_paren_is_all('(K1+K2-(K3+K5-K4))')
    assert not first_open_paren_is_all('(K1+K2-)K3+K5-K4)')


def test_split_into_steps():
    true_steps = ['K00330', 'K00331+K00332,K00331+K13378,K13380']
    assert split_into_steps('K00330+(K00331+K00332,K00331+K13378,K13380)', '+') == true_steps


def test_is_ko():
    assert is_ko('K00000')
    assert not is_ko('K1')


@pytest.fixture()
def module_network():
    network = nx.DiGraph()
    network.add_edges_from([('start', 'K00001'), ('K00001', 'K00002'), ('K00001', 'K00003'), ('K00002', 'K00004'),
                            ('K00003', 'K00013'), ('K00013', 'K00004')])
    return network


def test_make_module_network(module_network):
    test_network, _ = make_module_network('K00001+(K00002,K00003+K00013)+K00004')
    assert nx.is_isomorphic(test_network, module_network)


def test_get_module_coverage(module_network):
    module_network.add_edge('K00004', 'end')
    assert (3, 1, 1/3, {'K00001'}, {'K00002', 'K00004'}) == get_module_coverage(module_network, {'K00001'})
    assert (4, 1, 1/4, {'K00003'}, {'K00001', 'K00013', 'K00004'}) == get_module_coverage(module_network, {'K00003'})
    assert (3, 1, 1/3, {'K00002'}, {'K00001', 'K00004'}) == get_module_coverage(module_network, {'K00002', 'K99999'})


@pytest.fixture()
def etc_module_df():
    return pd.DataFrame([['K00001+(K00002,K00003+K00013)+K00004', 'Complex I', 'oxidoreductase', 'M00000']],
                        columns=['definition', 'complex', 'module_name', 'module_id'])


@pytest.fixture()
def etc_coverage_df():
    return pd.DataFrame([['M00000', 'oxidoreductase', 'I', 'scaffold_1', 3, 1, 1/3, 'K00001',
                          'K00002,K00004', 'Complex I: oxidoreductase']],
                        columns=['module_id', 'module_name', 'complex', 'genome', 'path_length',
                                 'path_length_coverage', 'percent_coverage', 'genes', 'missing_genes',
                                 'complex_module_name'])


def test_make_etc_coverage_df(test_annotations_df, etc_module_df, etc_coverage_df):
    test_etc_coverage_df = make_etc_coverage_df(etc_module_df, test_annotations_df, 'scaffold')
    pd.testing.assert_frame_equal(test_etc_coverage_df, etc_coverage_df)


def test_make_etc_coverage_heatmap(etc_coverage_df):
    hm = make_etc_coverage_heatmap(etc_coverage_df)
    assert type(hm) is alt.HConcatChart


@pytest.fixture()
def function_heatmap_form():
    return pd.DataFrame([['Category1', 'SubCategory1', 'A function', 'K00001, K99999', 'A long function name', ''],
                         ['Category1', 'SubCategory1', 'A function', 'K00002', 'A second long function name', ''],
                         ['Category1', 'SubCategory2', 'B function', 'K12345', 'A longer function name', '']],
                        columns=['category', 'subcategory', 'function_name', 'function_ids', 'long_function_name',
                                 'gene_symbol'])


@pytest.fixture()
def functional_df():
    return pd.DataFrame([['Category1', 'SubCategory1', 'A function', 'K00001',
                          'A long function name; A second long function name', '', 'scaffold_1', False,
                          'Category1: A function'],
                         ['Category1', 'SubCategory2', 'B function', 'K12345', 'A longer function name', '',
                          'scaffold_1', True, 'Category1: B function']],
                        columns=['category', 'subcategory', 'function_name', 'function_ids', 'long_function_name',
                                 'gene_symbol', 'genome', 'present', 'category_function_name'])


def test_make_functional_df(test_annotations_df, function_heatmap_form, functional_df):
    test_functional_df = make_functional_df(test_annotations_df, function_heatmap_form, 'scaffold')
    pd.testing.assert_frame_equal(test_functional_df, functional_df)


def test_make_functional_heatmap(functional_df):
    hm = make_functional_heatmap(functional_df)
    assert type(hm) is alt.HConcatChart


# TODO: actually test that the frames are correct, already done above
def test_fill_liquor_dfs(test_annotations_df, test_module_net, etc_module_df, function_heatmap_form):
    module_nets = {'M12345': test_module_net}
    liquor_dfs = fill_liquor_dfs(test_annotations_df, module_nets, etc_module_df, function_heatmap_form, 'scaffold')
    assert len(liquor_dfs) == 3
    assert type(liquor_dfs[0]) is pd.DataFrame
    assert type(liquor_dfs[1]) is pd.DataFrame
    assert type(liquor_dfs[2]) is pd.DataFrame


def test_make_liquor_heatmap(module_coverage_frame, etc_coverage_df, functional_df):
    hm = make_liquor_heatmap(module_coverage_frame, etc_coverage_df, functional_df)
    assert type(hm) is alt.HConcatChart


def test_make_liquor_df(module_coverage_frame, etc_coverage_df, functional_df):
    liquor_df = pd.DataFrame([[1/3, 1/3, False, True]],
                             index=pd.Index(['scaffold_1'], name='genome'),
                             columns=['a module name', 'Complex I: oxidoreductase', 'Category1: A function',
                                      'Category1: B function'])
    test_liquor_df = make_liquor_df(module_coverage_frame, etc_coverage_df, functional_df)
    pd.testing.assert_frame_equal(test_liquor_df, liquor_df)


def test_get_phylum_and_most_specific():
    assert get_phylum_and_most_specific('d__Bacteria;p__Bacteroidota;c__;o__;f__;g__;s__') == \
           'p__Bacteroidota;c__'
    assert get_phylum_and_most_specific('d__Archaea;p__;c__;o__;f__;g__;s__') == 'd__Archaea;p__'
    assert get_phylum_and_most_specific('d__Bacteria;p__Bacteroidota;c__Bacteroidia;o__Bacteroidales;f__Rikenellaceae;'
                                        'g__Alistipes;s__') == 'p__Bacteroidota;g__Alistipes'
    assert get_phylum_and_most_specific('d__Bacteria;p__Firmicutes;c__Bacilli;o__Lactobacillales;f__Enterococcaceae;'
                                        'g__Enterococcus_D;s__Enterococcus_D gallinarum') == \
        'p__Firmicutes;s__Enterococcus_D gallinarum'



import pandas as pd
import altair as alt
import re
from os import path, mkdir
from functools import partial
from collections import defaultdict, Counter
from datetime import datetime
import warnings

from mag_annotator.utils import get_database_locs
from mag_annotator.summarize_genomes import get_ordered_uniques

VOGDB_TYPE_NAMES = {'Xr': 'Viral replication genes', 'Xs': 'Viral structure genes',
                    'Xh': 'Viral genes with host benefits', 'Xp': 'Viral genes with viral benefits',
                    'Xu': 'Viral genes with unknown function', 'Xx': 'Viral hypothetical genes'}
VIRUS_STATS_COLUMNS = ['VIRSorter category', 'Circular', 'Prophage', 'Gene count', 'Strand switches',
                       'potential AMG count', 'Transposase present', 'Possible Non-Viral Contig']
VIRAL_DISTILLATE_COLUMNS = ['gene', 'scaffold', 'gene_id', 'gene_description', 'category', 'header',
                            'subheader', 'module', 'auxiliary_score', 'amg_flags']
VIRAL_LIQUOR_HEADERS = ['Category', 'Function', 'AMG Genes', 'Genes Present', 'Contig Name', 'Present in Contig']
HEATMAP_CELL_HEIGHT = 10
HEATMAP_CELL_WIDTH = 10

defaultdict_list = partial(defaultdict, list)

def filter_to_amgs(annotations, max_aux=4, remove_transposons=True, remove_fs=False):
    # def filter_to_amgs(annotations, max_aux=4, remove_transposons=True, remove_fs=False, remove_js=False):
    potential_amgs = list()
    for gene, row in annotations.iterrows():
        amg_flags = row['amg_flags']
        if not pd.isna(amg_flags):
            vmap_aux_check = ('V' not in amg_flags) and ('M' in amg_flags) and (row['auxiliary_score'] <= max_aux) and \
                             ('A' not in amg_flags) and ('P' not in amg_flags)
            remove_t = (remove_transposons and 'T' not in amg_flags) or not remove_transposons
            remove_f = (remove_fs and 'F' not in amg_flags) or not remove_fs
            # remove_j = (remove_fs and 'J' not in amg_flags) or not remove_js
            # if vmap_aux_check and remove_t and remove_f and remove_j:
            if vmap_aux_check and remove_t and remove_f:
                potential_amgs.append(gene)
    return annotations.loc[potential_amgs]


def get_strand_switches(strandedness):
    switches = 0
    strand = strandedness[0]
    for i in range(len(strandedness)):
        if strandedness[i] != strand:
            switches += 1
            strand = strandedness[i]
    return switches


def make_viral_stats_table(annotations, potential_amgs, groupby_column='scaffold'):
    amg_counts = potential_amgs.groupby(groupby_column).size()
    viral_stats_series = list()
    for scaffold, frame in annotations.groupby(groupby_column):
        # get virus information
        virus_category = int(re.findall(r'-cat_\d$', scaffold)[0].split('_')[-1])  # viral category
        virus_circular = len(re.findall(r'-circular-cat_\d$', scaffold)) == 1  # virus is circular
        virus_prophage = virus_category in [4, 5]  # virus is prophage
        virus_num_genes = len(frame)  # number of genes on viral contig
        virus_strand_switches = get_strand_switches(frame.strandedness)  # number of strand switches
        if scaffold in amg_counts:
            virus_number_amgs = amg_counts[scaffold]  # number of potential amgs
        else:
            virus_number_amgs = 0
        virus_transposase_present = sum(frame.is_transposon) > 0  # transposase on contig
        # virus_j_present = sum(['J' in i if not pd.isna(i) else False for i in frame.amg_flags]) > 0
        virus_j_present = sum([i == 'Xh' if not pd.isna(i) else False
                               for i in frame['vogdb_categories']]) / frame.shape[0]
        virus_data = pd.Series([virus_category, virus_circular, virus_prophage, virus_num_genes, virus_strand_switches,
                                virus_number_amgs, virus_transposase_present, virus_j_present],
                               index=VIRUS_STATS_COLUMNS, name=scaffold)
        # get vogdb categories
        # when vogdb has multiple categories only the first is taken
        gene_counts = Counter([i.split(';')[0] for i in frame.vogdb_categories.fillna('Xx')])
        named_gene_counts = {VOGDB_TYPE_NAMES[key]: value for key, value in gene_counts.items()}
        gene_counts_series = pd.Series(named_gene_counts, name=scaffold)
        viral_stats_series.append(virus_data.append(gene_counts_series))
    return pd.DataFrame(viral_stats_series).fillna(0)


def get_ids_from_row(row):
    id_list = list()
    # get kegg ids
    if 'kegg_id' in row and not pd.isna(row['kegg_id']):
        id_list += [j for j in row['kegg_id'].split(',')]
    # get ec numbers
    if 'kegg_hit' in row and not pd.isna(row['kegg_hit']):
        id_list += [i[1:-1] for i in re.findall(r'\[EC:\d*.\d*.\d*.\d*\]', row['kegg_hit'])]
    # get merops ids
    if 'peptidase_family' in row and not pd.isna(row['peptidase_family']):
        id_list += [j for j in row['peptidase_family'].split(';')]
    # get cazy ids
    if 'cazy_hits' in row and not pd.isna(row['cazy_hits']):
        id_list += [j.strip().split(' ')[0] for j in row['cazy_hits'].split(';')]
    return set(id_list)


def make_viral_distillate(potential_amgs, genome_summary_frame):
    rows = list()
    for gene, row in potential_amgs.iterrows():
        gene_ids = get_ids_from_row(row) & set(genome_summary_frame.index)
        if len(gene_ids) > 0:
            for gene_id in gene_ids:
                gene_summary = genome_summary_frame.loc[gene_id]
                if type(gene_summary) is pd.Series:
                    rows.append([gene, row['scaffold'], gene_id, gene_summary['gene_description'],
                                 gene_summary['sheet'], gene_summary['header'], gene_summary['subheader'],
                                 gene_summary['module'], row['auxiliary_score'], row['amg_flags']])
                else:
                    for sub_gene_id, sub_gene_summary in gene_summary.iterrows():
                        rows.append([gene, row['scaffold'], gene_id, sub_gene_summary['gene_description'],
                                     sub_gene_summary['sheet'], sub_gene_summary['header'],
                                     sub_gene_summary['subheader'], sub_gene_summary['module'],
                                     row['auxiliary_score'], row['amg_flags']])
        else:
            warnings.warn("No distillate information found for gene %s" % gene)
            rows.append([gene, row['scaffold'], '', '', '', '', '', '', row['auxiliary_score'],
                         row['amg_flags']])
    return pd.DataFrame(rows, columns=VIRAL_DISTILLATE_COLUMNS)


def make_vgf_order(amgs):
    amg_score_dict = {scaffold: ((1/frame['auxiliary_score']).sum(), len(frame))
                      for scaffold, frame in amgs.groupby('scaffold')}
    amg_scores = pd.DataFrame.from_dict(amg_score_dict, columns=['AMG_score', 'AMG_count'],
                                        orient='index')
    return list(amg_scores.sort_values(['AMG_score', 'AMG_count'], ascending=False).index)


def make_amg_count_column(potential_amgs, vgf_order=None):
    # build count column
    amg_counts = pd.DataFrame(Counter(potential_amgs.scaffold).items(), columns=['Contig Name', 'Number'])
    amg_counts['AMG Count'] = 'AMG Count'
    text = alt.Chart(amg_counts, width=HEATMAP_CELL_WIDTH+10, height=HEATMAP_CELL_HEIGHT*len(amg_counts)).encode(
                     x=alt.X('AMG Count', title=None, axis=alt.Axis(labelLimit=0, labelAngle=90)),
                     y=alt.Y('Contig Name', title=None, axis=alt.Axis(labelLimit=0), sort=vgf_order),
                     text='Number'
                    ).mark_text()
    return text


def make_viral_functional_df(annotations, genome_summary_frame, groupby_column='scaffold'):
    # build dict of ids per genome
    vgf_to_id_dict = defaultdict(defaultdict_list)
    for vgf, frame in annotations.groupby(groupby_column, sort=False):
        for gene, row in frame.iterrows():
            id_list = get_ids_from_row(row)
            for id_ in id_list:
                vgf_to_id_dict[vgf][id_].append(gene)
    # build long from data frame
    rows = list()
    for category, category_frame in genome_summary_frame.groupby('sheet'):
        for header, header_frame in category_frame.groupby('module'):
            header_id_set = set(header_frame.index.to_list())
            curr_rows = list()
            for vgf, id_dict in vgf_to_id_dict.items():
                present_in_bin = False
                functions_present = list()
                amgs_present = list()
                for id_, amgs in id_dict.items():
                    if id_ in header_id_set:
                        present_in_bin = True
                        functions_present.append(id_)
                        amgs_present += amgs
                curr_rows.append([category, header, ', '.join(amgs_present), ', '.join(functions_present), vgf,
                                  present_in_bin])
            if sum([i[-1] for i in curr_rows]) > 0:
                rows += curr_rows
    return pd.DataFrame(rows, columns=VIRAL_LIQUOR_HEADERS)


def make_viral_functional_heatmap(functional_df, vgf_order=None):
    # build heatmaps
    charts = list()
    for i, (group, frame) in enumerate(functional_df.groupby('Category', sort=False)):
        # set variables for chart
        function_order = get_ordered_uniques(list(frame['Function']))
        num_vgfs_in_frame = len(set(frame['Contig Name']))
        chart_width = HEATMAP_CELL_WIDTH * len(function_order)
        chart_height = HEATMAP_CELL_HEIGHT * num_vgfs_in_frame
        # set up colors for chart
        rect_colors = alt.Color('Present in Contig',
                                legend=alt.Legend(symbolType='square', values=[True, False]),
                                sort=[True, False],
                                scale=alt.Scale(range=['#e5f5f9', '#2ca25f']))
        # define chart
        # TODO: Figure out how to angle title to take up less space
        c = alt.Chart(frame, title=alt.TitleParams(group)).encode(
            x=alt.X('Function', title=None, axis=alt.Axis(labelLimit=0, labelAngle=90), sort=function_order),
            y=alt.Y('Contig Name', axis=alt.Axis(title=None, labels=False, ticks=False), sort=vgf_order),
            tooltip=[alt.Tooltip('Contig Name'),
                     alt.Tooltip('Category'),
                     alt.Tooltip('Function'),
                     alt.Tooltip('AMG Genes'),
                     alt.Tooltip('Genes Present')]
        ).mark_rect().encode(color=rect_colors).properties(
            width=chart_width,
            height=chart_height)
        charts.append(c)
    # merge and return
    function_heatmap = alt.hconcat(*charts, spacing=5)
    return function_heatmap


# def summarize_vgfs(input_file, output_dir, groupby_column='scaffold', max_auxiliary_score=3, remove_transposons=False,
#                    remove_fs=False, remove_js=False):
def summarize_vgfs(input_file, output_dir, groupby_column='scaffold', max_auxiliary_score=3,
                   remove_transposons=False, remove_fs=False):
    start_time = datetime.now()

    # set up
    annotations = pd.read_csv(input_file, sep='\t', index_col=0)
    db_locs = get_database_locs()
    if 'genome_summary_form' not in db_locs:
        raise ValueError('Genome summary form location must be set in order to summarize genomes')
    mkdir(output_dir)
    genome_summary_form = pd.read_csv(db_locs['genome_summary_form'], sep='\t', index_col=0)
    print('%s: Retrieved database locations and descriptions' % (str(datetime.now() - start_time)))

    # get potential AMGs
    # potential_amgs = filter_to_amgs(annotations.fillna(''), max_aux=max_auxiliary_score,
    #                                 remove_transposons=remove_transposons, remove_fs=remove_fs, remove_js=remove_js)
    potential_amgs = filter_to_amgs(annotations.fillna(''), max_aux=max_auxiliary_score,
                                    remove_transposons=remove_transposons, remove_fs=remove_fs)
    print('%s: Determined potential amgs' % (str(datetime.now() - start_time)))

    # make distillate
    viral_genome_stats = make_viral_stats_table(annotations, potential_amgs, groupby_column)
    viral_genome_stats.to_csv(path.join(output_dir, 'vMAG_stats.tsv'), sep='\t')
    print('%s: Calculated viral genome statistics' % (str(datetime.now() - start_time)))

    viral_distillate = make_viral_distillate(potential_amgs, genome_summary_form)
    viral_distillate.to_csv(path.join(output_dir, 'amg_summary.tsv'), sep='\t', index=None)
    print('%s: Generated AMG summary' % (str(datetime.now() - start_time)))

    # make liquor
    vgf_order = make_vgf_order(potential_amgs)
    amg_column = make_amg_count_column(potential_amgs, vgf_order)
    viral_function_df = make_viral_functional_df(potential_amgs, genome_summary_form, groupby_column=groupby_column)
    viral_functional_heatmap = make_viral_functional_heatmap(viral_function_df, vgf_order)
    alt.hconcat(amg_column, viral_functional_heatmap, spacing=5).save(path.join(output_dir, 'product.html'))
    print('%s: Generated product heatmap' % (str(datetime.now() - start_time)))
    print("%s: Completed distillation" % str(datetime.now() - start_time))



from os import path
from warnings import warn

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from mag_annotator.database_setup import TABLE_NAME_TO_CLASS_DICT
from mag_annotator.utils import divide_chunks


# TODO: store all sequence db locs within database handler class


class DatabaseHandler:
    def __init__(self, database_loc):
        if not path.exists(database_loc):
            raise ValueError('Database does not exist at path %s' % database_loc)
        engine = create_engine('sqlite:///%s' % database_loc)
        db_session = sessionmaker(bind=engine)
        self.session = db_session()

    # functions for adding descriptions to tables
    def add_descriptions_to_database(self, description_list, db_name, clear_table=True):
        description_class = TABLE_NAME_TO_CLASS_DICT[db_name]
        if clear_table:
            self.session.query(description_class).delete()
        self.session.bulk_save_objects([description_class(**i) for i in description_list])
        self.session.commit()
        self.session.expunge_all()

    # functions for getting descriptions from tables
    def get_description(self, annotation_id, db_name):
        return self.session.query(TABLE_NAME_TO_CLASS_DICT[db_name]).filter_by(id=annotation_id).one().description

    def get_descriptions(self, ids, db_name):
        description_class = TABLE_NAME_TO_CLASS_DICT[db_name]
        descriptions = list()
        for chunk in divide_chunks(list(ids), 499):
            descriptions += self.session.query(description_class).filter(description_class.id.in_(chunk)).all()

        if len(descriptions) == 0:
            warn("No descriptions were found for your id's. Does this %s look like an id from %s" % (list(ids)[0],
                                                                                                     db_name))
        return {i.id: i.description for i in descriptions}

    @staticmethod
    def get_database_names():
        return TABLE_NAME_TO_CLASS_DICT.keys()



from setuptools import setup, find_packages
from mag_annotator import __version__ as version
from os import path

__author__ = 'shafferm'
__version__ = version

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="DRAM-bio",
    version=__version__,
    scripts=['scripts/DRAM.py', 'scripts/DRAM-v.py', 'scripts/DRAM-setup.py'],
    packages=find_packages(),
    description="Distilled and Refined Annotation of Metabolism: A tool for the annotation and curation of function for"
                " microbial and viral genomes",
    long_description=long_description,
    long_description_content_type='text/markdown',  # Optional (see note above)
    package_data={'mag_annotator': ['CONFIG']},
    python_requires='>=3',
    install_requires=['scikit-bio', 'pandas', 'altair', 'sqlalchemy', 'networkx', 'openpyxl', 'numpy'],
    author="Michael Shaffer",
    author_email='michael.t.shaffer@colostate.edu',
    url="https://github.com/shafferm/DRAM/",
    download_url="https://github.com/shafferm/DRAM/tarball/%s" % __version__
)
